/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	BASCNV:					*/
/*	Define all routine to manage language	*/
/*	string conversion.			*/
/*						*/
/************************************************/
#include	<string.h>
#include	<ctype.h>
#include	<iconv.h>

#define		OPENSSL_NO_KRB5
#include	<openssl/bio.h>
#include	<openssl/evp.h>
#include	<openssl/err.h>
#include	<openssl/ssl.h>

#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"bascnv.h"

#define	DMRK	"=?"
#define	EMRK	"?="
#define	DLANG	"UNICODE"

typedef	struct	{
	char *lang;	/*coding language	*/
	char encoding;	/*encoding sequence type*/
	char *encoded;	/*encoded text		*/
	}QTYP;
/*

*/
/************************************************/
/*						*/
/*	Routine tocheck if all char within a B64*/
/*	string are indeed B6$			*/
/*						*/
/************************************************/
static int checkb64(char *b64)

{
int ok;

ok=true;
while ((*b64!='\000')&&(ok==true)) {
  switch (*b64)	{
    case '+'	:
    case '/'	:
    case '='	:
      break;
    default	:
      if (isalnum(*b64)==0) {
	ok=false;
	*b64='\000';
	}
      break;
    }
  b64++;
  }
return ok;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to free a QTYP struct		*/
/*						*/
/************************************************/
static QTYP *freeqtyp(QTYP *qtyp)

{
if (qtyp !=(QTYP *)0) {
  if (qtyp->encoded!=(char *)0)
    (void) free(qtyp->encoded);
  if (qtyp->lang!=(char *)0)
    (void) free(qtyp->lang);
  (void) free(qtyp);
  qtyp=(QTYP *)0;
  }
return qtyp;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to creat a QTYP from a quoted	*/
/*	printable sequence			*/
/*	contains within a string		*/
/*						*/
/************************************************/
static QTYP *mkqtyp(char **quoted)

{
QTYP *qtyp;
char **split;
int phase;
int proceed;

qtyp=(QTYP *)0;
split=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*lets split quoted	*/
      if (*quoted!=(char *)0) { /*always	*/
	int i;

	for (i=0;i<3;i++) {
	  char *ptr;

	  if ((ptr=strchr(*quoted,'?'))!=(char *)0) {
	    *ptr='\000';
	    
	    split=(char **)vec_addlstlst((void **)split,(void *)strdup(*quoted));  
	    *quoted=ptr+1;
	    }
	  }
	if ((i!=3)||(**quoted!='='))
	  phase=999;	/*real trouble		*/
	}
      break;
    case 1	:	/*we got 3 part		*/
    			/*making qtype		*/
      qtyp=(QTYP *)calloc(1,sizeof(QTYP)); 
      qtyp->lang=strdup(split[0]);
      qtyp->encoding=tolower(split[1][0]);
      qtyp->encoded=strdup(split[2]);
      *quoted+=1;	/*skipping '=?' (EMRK)	*/
      if (**quoted==' ')
	*quoted+=1;	/*skipping next space	*/
	
      break;
    default	:	/*SAFE Guard		*/
      split=(char **)vec_freelstlst((void **)split,VECFREE);
      proceed=false;
      break;
    }
  phase++;
  }
return qtyp;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to split MIME quoted-printable	*/
/*	contains within a string		*/
/*						*/
/************************************************/
static QTYP **splitmime(char *mimeq)

{
QTYP **qtyps;

qtyps=(QTYP **)0;
while (*mimeq!='\000') {
  QTYP *qloc;

  qloc=(QTYP *)0;
  if (strstr(mimeq,DMRK)==mimeq) {
    mimeq+=strlen(DMRK);
    qloc=mkqtyp(&mimeq);
    }
  else {
    char *ptr;

    ptr=(char *)0;
    qloc=(QTYP *)calloc(1,sizeof(QTYP)); 
    qloc->encoded=(char *)calloc(strlen(mimeq)+2,sizeof(char));
    (void) strcpy(qloc->encoded," ");
    (void) strcat(qloc->encoded,mimeq);
    (void) strcpy(mimeq,"");
    if ((ptr=strstr(qloc->encoded,DMRK))!=(char *)0) {
      (void) strcpy(mimeq,ptr);
      *ptr='\000';
      }
    }
  if (qloc!=(QTYP *)0) {
    qtyps=(QTYP **)vec_addlstlst((void **)qtyps,qloc);
    }
  }
return qtyps;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert an char sequence	*/
/*	from on char set to another.		*/
/*	if successful return a new pointer	*/
/*						*/
/************************************************/
static char *cnvconvert(char *lfrom,char *lto,char *encoded)

{
#define	OPCN	"bascnv.c:cnvconvert,"

char *converted;
char *marker;
iconv_t hconv;
size_t inbuflft;
char *inbuf;
size_t outbuflft;
char *outbuf;
int phase;
int proceed;

converted=(char *)0;
marker=(char *)0;
hconv=(iconv_t)0;
inbuflft=(size_t)0;
inbuf=(char *)0;
outbuflft=(size_t)0;
outbuf=(char *)0;
phase=0;
proceed=1;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening conv channel	*/
      if ((hconv=iconv_open(lto,lfrom))==(iconv_t)-1) {
	(void) clm_alert(0,"%s Unexpected lang conversion request from=<%s> to=<%s> (errno=<%s>)",
			    OPCN,lfrom,lto,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*doing convertion	*/
      inbuflft=strlen(encoded);
      outbuflft=(inbuflft*4)+1;	/*spare room	*/
      inbuf=encoded;
      marker=(char *)calloc(outbuflft,sizeof(char));
      outbuf=marker;
      if (iconv(hconv,&inbuf,&inbuflft,&outbuf,&outbuflft)==(size_t)(-1)) {
	(void) clm_alert(2,"%s Unable to convert <%s> from=<%s> to=<%s> (errno=<%s>)",
			    OPCN,encoded,lfrom,lto,strerror(errno));
	(void) free(marker);
	marker=(char *)0;
	}
      break;
    case 2	:	/*doing convertion	*/
      if (iconv_close(hconv)<0) 
	(void) clm_alert(0,"%s Unexpected close conversion error (errno=<%s>) (bug?)",
			    OPCN,strerror(errno));
      break;
    default	:	/*SAFE Guard		*/
      converted=marker;
      proceed=false;
      break;
    }
  phase++;
  }
return converted;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to convert a quoted-printable	*/
/*	sequence to proper destination.		*/
/*						*/
/************************************************/
static char *mimeconv(char *tolang,QTYP *qtyp)

{
#define	OPMI	"unicnv.c:mimeconv,"

char *conv;
char *result;
int phase;
int proceed;

conv=(char *)0;
result=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*qtyp avail		*/
      if (qtyp==(QTYP *)0)
	phase=999;	/*trouble trouble	*/
      break;
    case 1	:	/*standard translation	*/
      if (qtyp->lang==(char *)0) {
	conv=qtyp->encoded;
	qtyp->encoded=(char *)0;
	phase=999;	/*that's all		*/
	}
      break;
    case 2	:	/*qtyp avail		*/
      switch(qtyp->encoding) {
	case 'b':
	  conv=cnv_getb64(qtyp->encoded);
	  break;
	case 'q':
	  conv=cnv_getqp(qtyp->encoded);
	  break;
	default	:
	  (void) clm_alert(0,"%s Unexpected qtyp encode '%c' (bug?)",OPMI,qtyp->encoding);
	  break;
	}
      if (conv==(char *)0)
	phase=999;	/*trouble trouble	*/
      break;
    case 3	:	/*doing file convertion	*/
      if ((result=cnvconvert(qtyp->lang,tolang,conv))!=(char *)0) {
        (void) free(conv);
	conv=result;
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return conv;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert an ASCII Quoted	*/
/*	printable Sequence to a plain 8bit	*/
/*	sequence.				*/
/*						*/
/************************************************/
char *cnv_getqp(const char *qp)

{
char *str;

str=(char *)0;
if (qp!=(char *)0) {
  register int i,j;
  char code[6];

  (void) strcpy(code,"0x00");
  str=calloc(strlen(qp)+3,sizeof(char));
  for (i=0,j=0;qp[i]!='\000';i++,j++) {
    switch (qp[i]) {
      case '_'	:
	str[j]=' ';
	break;
      case '='	:
        (void) memmove(code+2,qp+i+1,2);
	str[j]=(char)strtol(code,(char **)0,16);	
	i+=2;
	break;
      default	:
	str[j]=qp[i];
	break;
      }
    }
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert an ASCII B64	*/
/*	Sequence to a plain ASCII sequence.	*/
/*						*/
/************************************************/
char *cnv_getb64(char *b64)

{
#define	OPB6	"unicnv.c:cnv_getb64,"

char *decoded;
BIO *mbio;
BIO *bbio;
int retour;

(void) checkb64(b64);
decoded=(char *)calloc(strlen(b64)+3,sizeof(char));
mbio=BIO_new(BIO_s_mem());
bbio=BIO_new(BIO_f_base64());
BIO_set_flags(bbio,BIO_FLAGS_BASE64_NO_NL);
bbio=BIO_push(bbio,mbio);
(void) BIO_write(mbio,b64,strlen(b64));
(void) BIO_flush(mbio);
switch(retour=BIO_read(bbio,decoded,strlen(b64))) {
  case -2	:
    (void) clm_alert(0,"%s Can't proceed! decoding <%s> (bug?)",OPB6,b64);
    break;
  case -1	:
  case  0	:
    (void) clm_alert(0,"%s Need a retry! decoding <%s> (bug?)",OPB6,b64);
    break;
  default	:
    if (retour>0) {	/*always		*/
      register int scan;
      register int skip;

      scan=retour;
      skip=strlen(IOBNULL);	
      while (scan>0) {
        scan--;
        if (decoded[scan]=='\000') {
          register int taille;

          taille=retour+skip+1;
	  decoded=(char *)realloc(decoded,taille*sizeof(char));
	  (void) memmove(decoded+scan+skip,decoded+scan+1,(retour-scan)+1);
          (void) memmove(decoded+scan,IOBNULL,skip);
	  retour+=skip;
	  }
	}
      }
    break;
  }
if (retour<=0) {
  (void) free(decoded);
  decoded=(char *)0;
  }
(void) BIO_free_all(bbio);
return decoded;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert a plain ASCII	*/
/*	sequence to an ASCII B64.		*/
/*						*/
/************************************************/
char *cnv_setb64(const char *str)

{
#define	OPBS	"unicnv.c:tls_setb64,"

register int skip;

char *coded;
BIO *bmem;
BIO *b64;
BUF_MEM *bptr;
int taille;
char *buffer;
char *ptr;

skip=strlen(IOBNULL)-1;
coded=(char *)0;
b64=BIO_new(BIO_f_base64());
BIO_set_flags(b64,BIO_FLAGS_BASE64_NO_NL);
bmem=BIO_new(BIO_s_mem());
b64=BIO_push(b64,bmem);
buffer=strdup(str);
taille=strlen(buffer);
ptr=buffer;
while ((ptr=strstr(ptr,IOBNULL))!=(char *)0) {
  *ptr='\000';
  ptr++;
  (void) memmove(ptr,ptr+skip,strlen(ptr+skip)+1);
  taille-=skip;
  }
(void) BIO_write(b64,buffer,taille);
(void) BIO_flush(b64);
BIO_get_mem_ptr(b64,&bptr);
(void) free(buffer);
coded=(char *)calloc(bptr->length+1,sizeof(char));
(void) memcpy(coded,bptr->data,bptr->length);
(void) BIO_free_all(b64);
return coded;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to convert a string in 		*/
/*	MIME					*/
/*	(Multipurpose Internet Mail Extensions)	*/
/*	to the local encoding format.		*/
/*						*/
/*	This routine is to convert encoding	*/
/*	Within Email header			*/
/*						*/
/************************************************/
char *cnv_mimequoted(char *strtoconv)

{
char *conv;
char *ptr;
char *lang;
QTYP **qtyps;
int phase;
int proceed;

conv=(char *)0;
ptr=(char *)0;
lang=(char *)0;
qtyps=(QTYP **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*something to convert	*/
      if (strtoconv==(char *)0) 
	phase=999;	/*nothing strange...	*/
      break;
    case 1	:	/*is there any MIME	*/
      if (strstr(strtoconv,DMRK)==(char *)0) {
        conv=strdup(strtoconv);
	phase=999;	/*clean cut		*/
	}
      break;
    case 2	:	/*using current lang	*/
      if (((ptr=getenv("DB_LANG"))==(char *)0)||(strlen(ptr)==0))
        ptr=DLANG;
      lang=strdup(ptr);
      if ((ptr=strrchr(lang,'.'))!=(char *)0) {
	ptr++;	
	(void) memmove(lang,ptr,strlen(ptr)+1);
	}
      conv=strdup("");
      break;
    case 3	:	/*using current lang	*/
      if ((qtyps=splitmime(strtoconv))!=(QTYP **)0) {
	int i;

	for (i=0;qtyps[i]!=(QTYP *)0;i++) {
	  char *dmime;

	  if ((dmime=mimeconv(lang,qtyps[i]))!=(char *)0) {
	    conv=(char *)realloc((void *)conv,(strlen(conv)+strlen(dmime)+2)*sizeof(char));
	    (void) strcat(conv,dmime);
	    (void) free(dmime);
	    }
	  qtyps[i]=freeqtyp(qtyps[i]);
	  }
	(void) free(qtyps);
	}
      (void) free(lang);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return conv;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to check if string is properly	*/
/*	code, if it is not the case replace all	*/
/*	dubious charactere with plain ASCII	*/
/*						*/
/************************************************/
void cnv_checkencoding(char *strencoded)

{
char *ptr;
int phase;
int proceed;

ptr=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*autoconvert string	*/
      if (((ptr=getenv("DB_LANG"))==(char *)0)||(strlen(ptr)==0))
        ptr=DLANG;
      if ((ptr=cnvconvert(ptr,ptr,strencoded))!=(char *)0) {
        (void) free(ptr);
        phase=999;	/*everything fine	*/
        }
      break;
    case 1	:	/*setting to plain ASCII*/
      if (strencoded!=(char *)0) { /*alway	*/
	while (*strencoded!='\000') {
	  if (isascii(*strencoded)==0)
	    *strencoded='_';
	  strencoded++;
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
