/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESDBC:					*/
/*	Take Care of all clement data-base	*/
/*	access and management.			*/
/*						*/
/************************************************/
#include	<string.h>
#include	<errno.h>
#include	<math.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unidns.h"
#include	"unieml.h"
#include	"unijrl.h"
#include	"gesdbc.h"

#define	BINDING	"binding"	/*binding table	*/
#define	RLYTBL	"relay"		/*relaying table*/
#define	SRVTBL	"services"	/*services table*/
#define	CLMGRP	"clmgroup"	/*clmgroup table*/
#define	CLMBIND	"clmbind"	/*clmbind table	*/
#define	DOMS	"domain"	/*domain table	*/
#define	USERS	"users"		/*users table	*/
#define	INTUSR	"intusers"	/*internal users*/
#define	ALIASES	"aliases"	/*aliases table	*/
#define	BLSHST	"blshosts"	/*BLS servers	*/
#define	EMLOGS	"emlogs"	/*sessions logs	*/
#define	EMLTRT	"emltrt"	/*transit logs	*/
#define	USEPRF	"userprofile"	/*usage profile	*/
#define	OVRRIDE	"overrides"	/*overriding set*/
#define	EXTRACT	"extract"	/*extract list	*/
#define	INQUEUE	"inqueue"	/*transit queue	*/
#define	DOMAL	"domainalias"	/*dom alias list*/
#define	SESLOGS	"seslogs"	/*session logs	*/
#define	CRON	"cron"		/*session logs	*/
#define BANTBL	"banned"	/*banned ip tbl	*/
#define TRSTBL	"trusted"	/*trusted ip tbl*/
#define RATEBAN	"ratebantrust"	/*rmtip rate	*/
#define MPATH	"multipath"	/*multipath tbl	*/

#define	SELFROM	"SELECT * FROM"
#define	SELDOM	"%s %s WHERE domainname=%s"

#define	DFLTLNG	1		/*default langue*/

#define	MAXINFO	8000	/*info field max length	*/

/*

*/
/************************************************/
/*						*/
/*	Procedure to convert RBANNED/RTRUSTED	*/
/*	status to the DB bantrust table value	*/
/*						*/
/************************************************/
static int getbanval(int status,char *routine)

{
int dbstatus;

dbstatus=0;
switch (status) {
  case RBANNED	:
    dbstatus=BANBAN;	/*banned		*/
    break;
  default	:
    (void) clm_alert(0,"%s Unexpected status='%d' (bug? from <%s>)","gesdbc.c:getbanval,",status,routine);
    break;
  }
return dbstatus;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage active status	*/
/*	return an strcmp value.			*/
/*						*/
/************************************************/
static int mngactive(char *username,int active)

#define	OPMN	"gesdbc.c:mngactive,"

{
switch (active) {
  case 0	:	/*user is not active	*/
  case 1	:	/*user is active	*/
  case 2	:	/*to rebounce group	*/
    break;		/*nothing to do		*/
  case 3	:	/*deleted user		*/
    active=0;		/*set user unactive	*/
    break;
  default	:	/*unexpect case		*/
    (void) clm_alert(0,"%s Unexpected active level '%d' for user <%s> within '%s' table!",
			OPMN,active,username,USERS);
    active=0;		/*set user unactive	*/
    break;
  }
return active;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare 2 aliases and	*/
/*	return an strcmp value.			*/
/*						*/
/************************************************/
static int cmpali(const void *p1,const void *p2)

{
return strcmp(*(char *const *)p1,*(char *const *) p2);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get an 'upto' date		*/
/*	including the grace period		*/
/*						*/
/************************************************/
static time_t getuptograce(SQLFILE *sql,SQLRESULT *rs,int recpos)

{
time_t upto;

upto=sql_tounixtime(sql,sql_getvalue(sql,rs,recpos,"upto"));
upto+=atoi(sql_getvalue(sql,rs,recpos,"grace"))*24*3600;
return upto;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to make the username search	*/
/*	string.					*/
/*						*/
/************************************************/
static char *getnsearch(SQLFILE *sql,char *rcpt)

{
char *nsearch;
char *username;

username=strdup(rcpt);
(void) eml_splitdomain(username);
if (strlen(username)>0) {
  char *cleanstr;

  cleanstr=sql_cleanstr(sql,username);
  (void) asprintf(&nsearch,"=%s",cleanstr);
  (void) free(cleanstr);
  }
else
  (void) asprintf(&nsearch,"%s","IS NULL");
(void) free(username);
return nsearch;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to make the domain search	*/
/*	string.					*/
/*						*/
/************************************************/
static char *getdsearch(char *name)

{
char *dsearch;
char *domain;

dsearch=(char *)0;
if ((domain=eml_getdomain(name))!=(char *)0) 
  (void) asprintf(&dsearch,"='%s'",domain);
else
  (void) asprintf(&dsearch,"%s","IS NULL");
return dsearch;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return an ordered list	*/
/*	of bantrust ip according net type, mask	*/
/*	and status.				*/
/*						*/
/************************************************/
static char **listbanned(SQLFILE *sql,int netype,int mask,int offset,int maxlst)

{
#define	SOLBAN	"SELECT ipnum FROM %s WHERE mask=%d AND iptype=%d ORDER BY ipbin ASC LIMIT %d OFFSET %d"

char **iplists;
SQLRESULT *rs;

iplists=(char **)0;
if ((rs=sql_tupple(sql,SOLBAN,BANTBL,mask,netype,maxlst,offset))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    register int i;

    for (i=0;i<nbr;i++) 
      iplists=(char **)vec_addlstlst((void **)iplists,strdup(sql_getvalue(sql,rs,i,"ipnum")));
    }
  (void) sql_dropresult(sql,rs);
  }
return iplists;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve group ID from	*/
/*	domain name, if the domain is unknown	*/
/*	return 0.				*/
/*						*/
/************************************************/
static u_int getgroupid(SQLFILE *sql,char *email)

{
u_int groupid;
char *ptrdom;
char *domain;
SQLRESULT *rs;

groupid=0;
ptrdom=eml_getdomain(email);
domain=sql_cleanstr(sql,ptrdom);
if ((ptrdom=dbc_isitadomainalias(sql,ptrdom))!=(char *)0) {
  (void) free(domain);
  domain=sql_cleanstr(sql,ptrdom);
  (void) free(ptrdom);
  }
if ((rs=sql_tupple(sql,SELDOM,SELFROM,DOMS,domain))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0)
    groupid=(atoi)(sql_getvalue(sql,rs,0,"groupid"));
  (void) sql_dropresult(sql,rs);
  }
(void) free(domain);
return groupid;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve the number of	*/
/*	email sent to outside world by an	*/
/*	authenticated user.			*/
/*						*/
/************************************************/
static u_int getcount(SQLFILE *sql,char *authemail,u_int grpid)

{
#define	SELCNT	\
   "SELECT count(*) FROM %s WHERE sender='%s' AND status=%d AND grpin=%d AND grpout!=%d AND stamp>'%s'"

register const char *YDAY;

u_int count;
SQLRESULT *rs;

count=0;
YDAY=sql_fromunixtime(sql,time((time_t *)0)-H24);
if ((rs=sql_tupple(sql,SELCNT,EMLOGS,authemail,m_delivered,grpid,grpid,YDAY))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0)
    count=(atoi)(sql_getfield(sql,rs,0,0));
  (void) sql_dropresult(sql,rs);
  }
return count;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve the emlog status	*/
/*	for a specific session+recipient.	*/
/*	If record is not found, return status	*/
/*	m_undefined.				*/
/*						*/
/************************************************/
static MENTYP getemlogstatus(SQLFILE *sql,char *session,char *rcpt)

{
#define	SELSTA	"SELECT status FROM %s WHERE session='%s' AND recipient=%s"

MENTYP status;
SQLRESULT *rs;

status=m_undefined;
if ((rs=sql_tupple(sql,SELSTA,EMLOGS,session,rcpt))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0)
    status=(atoi)(sql_getvalue(sql,rs,0,"status"));
  (void) sql_dropresult(sql,rs);
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to insert a queue entry within*/
/*	the table 'inqueue'			*/
/*	return "true" if not trouble, false	*/
/*	otherwise.				*/
/*	available.				*/
/*						*/
/************************************************/
static int setqueue(SQLFILE *sql,QUETYP *q)

{
#define	OPSQ	"gesdbc.c:setqueue,"

#define INSFLD1	"session,transit,outip,smtpmx,smtpport,subject,rejectto,orgeml,orgdom"
#define INSFLD2	"grpin,recname,recdom,mode,start,last,next,upto,orgstatus"
#define	INSQUE	\
  "INSERT INTO %s (%s,%s) VALUES('%s',%d,'%s','%s',%d,%s,%s,%s,'%s',%d,%s,'%s',%d,NOW(),NOW(),NOW(),'%s',%d)"

int status;
char *subject;
char *rejectto;
char *orgeml;
char *receml;
register const char *TWODAY;

status=true;
subject=sql_cleanstr(sql,q->subject);
rejectto=sql_cleanstr(sql,q->rejectto);
orgeml=sql_cleanstr(sql,q->orgeml);
receml=sql_cleanstr(sql,q->recname);
TWODAY=sql_fromunixtime(sql,time((time_t *)0)+(2*H24));
if (sql_action(sql,INSQUE,INQUEUE,INSFLD1,INSFLD2,q->session,q->transit,
					 	  q->outip,q->smtpmx,atoi(q->smtpport),subject,
					 	  rejectto,orgeml,eml_getdomain(q->orgeml),q->grpin,
					 	  receml,q->recdom,(int)q->mode,TWODAY,q->orgstatus)!=1) {
  (void) clm_alert(0,"%s Unable store session-iteration <%s-%03d> (BUG?)",
		      OPSQ,q->session,q->transit);
  status=false;
  }
(void) free(receml);
(void) free(orgeml);
(void) free(rejectto);
(void) free(subject);
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract One Queue from	*/
/*	data-base.				*/
/*						*/
/************************************************/
static QUETYP *getonequeue(SQLFILE *sql,SQLRESULT *rs,int pos)

{
QUETYP *queue;
register char *session;
register char *recname;
register char *recdom;
register MMOENU mode;
register int orgstatus;
register char *outip;
register char *smtpmx;
register char *smtpport;
register u_long transit;
register char *subject;
register char *rejectto;
register char *orgeml;
register int grpin;

session=sql_getvalue(sql,rs,pos,"session");
recname=sql_getvalue(sql,rs,pos,"recname");
recdom=sql_getvalue(sql,rs,pos,"recdom");
mode=(MMOENU)atoi(sql_getvalue(sql,rs,pos,"mode"));
orgstatus=atoi(sql_getvalue(sql,rs,pos,"orgstatus"));
outip=sql_getvalue(sql,rs,pos,"outip");
smtpmx=sql_getvalue(sql,rs,pos,"smtpmx");
smtpport=sql_getvalue(sql,rs,pos,"smtpport");
transit=atol(sql_getvalue(sql,rs,pos,"transit"));
subject=sql_getvalue(sql,rs,pos,"subject");
rejectto=sql_getvalue(sql,rs,pos,"rejectto");
orgeml=sql_getvalue(sql,rs,pos,"orgeml");
grpin=atoi(sql_getvalue(sql,rs,pos,"grpin"));
queue=que_mkqueue(outip,smtpmx,smtpport,session,transit,subject,rejectto,orgeml,grpin,recname,recdom,mode,orgstatus);
queue->start=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"start"));
queue->upto=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"upto"));
queue->nexttime=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"next"));
queue->lasttime=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"last"));
queue->retry=atoi(sql_getvalue(sql,rs,pos,"retry"));
queue->mode=atoi(sql_getvalue(sql,rs,pos,"mode"));
queue->numtrace=atol(sql_getvalue(sql,rs,pos,"numtrace"));
return queue;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a list of queued	*/
/*	email ready to be processed.		*/
/*						*/
/************************************************/
static QUETYP **getpendingqueue(SQLFILE *sql,int max)

{
#define SELQUE	"SELECT * FROM %s WHERE next<='%s' AND progress=%d ORDER BY recdom,smtpmx,session LIMIT %d FOR UPDATE"
QUETYP **queues;
SQLRESULT *rs;
register time_t limite;

queues=(QUETYP **)0;
limite=time((time_t *)0);
if ((rs=sql_tupple(sql,SELQUE,INQUEUE,sql_fromunixtime(sql,limite),QWAIT,max))!=(SQLRESULT *)0) {
  int nbr;
  int i;

  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    register char *session;
    register char *recname;
    register char *recdom;
    register MMOENU mode;
    	     char *recipient;
    register char *rcpt;
   
     
    session=sql_getvalue(sql,rs,i,"session");
    recname=sql_getvalue(sql,rs,i,"recname");
    recdom=sql_getvalue(sql,rs,i,"recdom");
    mode=(MMOENU)atoi(sql_getvalue(sql,rs,i,"mode"));
    (void) asprintf(&recipient,"%s@%s",recname,recdom);
    rcpt=sql_cleanstr(sql,recipient);
    if ((mode!=mm_std)||(getemlogstatus(sql,session,rcpt)!=m_undefined)) {
      QUETYP *q;

      q=getonequeue(sql,rs,i);
      queues=(QUETYP **)vec_addlstlst((void **)queues,(void *)q);
      }
    (void) free(rcpt);
    (void) free(recipient);
    }
  (void) sql_dropresult(sql,rs);
  }
return queues;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the alias list if	*/
/*	available.				*/
/*						*/
/************************************************/
static char **getalias(SQLFILE *sql,char *rcpt,char ***known)

{
#define	OPGT	"gesdbc.c:getalias,"

#define	SELALI	"%s %s WHERE domain %s AND username %s"

char **aliases;
int phase;
int proceed;
char *username;
char *domain;
SQLRESULT *rs;

aliases=(char **)0;
phase=0;
proceed=true;
username=getnsearch(sql,rcpt);
domain=getdsearch(rcpt);
rs=(SQLRESULT *)0;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*aliase already seen	*/
      if (*known!=(char **)0) {
	int i;

	for (i=0;(*known)[i]!=(char *)0;i++) {
	  if (strcmp((*known)[i],rcpt)==0) {
	    aliases=(char **)calloc(1,sizeof(char *));
	    phase=999;
	    break;
	    }
	  }
	}
      break;
    case 1	:	/*extracting aliase	*/
      if ((rs=sql_tupple(sql,SELALI,SELFROM,ALIASES,domain,username))!=(SQLRESULT *)0) {
	if (sql_nbrtupple(sql,rs)>0) {
	  aliases=clm_splitstring(sql_getvalue(sql,rs,0,"aliasnames"),',');
	  if (aliases==(char **)0)
	  /*to mark is an empty alias!		*/
	    aliases=(char **)calloc(1,sizeof(char *));
          *known=(char **)vec_addlstlst((void **)(*known),(void *)strdup(rcpt));
	  }
        (void) sql_dropresult(sql,rs);
  	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(domain);
(void) free(username);
return aliases;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve aliase list	*/
/*						*/
/************************************************/
static char **getaliaslist(SQLFILE *sql,char *rcpt,char ***known,int depth)

{
#define	OPGL	"gesdbc.c:getaliaslist,"
#define	MXADPTH	10

char **aliases;
char **sup;
int phase;
int proceed;

aliases=(char **)0;
sup=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase)  {
    case 0	:	/*critical depth reach!	*/
      if (depth>=MXADPTH) {
	(void) clm_alert(0,"%s, too many aliases recursions '%d>=%d'",OPGL,depth,MXADPTH);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*let get the alias list*/
      if ((sup=getalias(sql,rcpt,known))==(char **)0) {
	phase=999;	/*it was not an alias	*/
	}
      break;
    case 2	:	/*is alias empty?	*/
      if (sup[0]==(char *)0) {
	aliases=sup;
	phase=999;	/*alias empty		*/
	}
      break;
    case 3	:	/*lets scan alias list	*/
      if (sup!=(char **)0) {	/*always	*/
	int i;

	for (i=0;sup[i]!=(char *)0;i++) {
	  char **subalias;

	  subalias=(char **)0;
	  if (strcmp(sup[i],rcpt)==0) {
	    aliases=(char **)vec_addlstlst((void **)aliases,(void *)sup[i]);
	    continue;
	    }
	  if ((subalias=getaliaslist(sql,sup[i],known,depth+1))!=(char **)0) {
	    int j;

	    for (j=0;subalias[j]!=(char *)0;j++) {
	      aliases=(char **)vec_addlstlst((void **)aliases,(void *)subalias[j]);
	      }
	    aliases=(char **)vec_addlstlst((void **)aliases,(void *)0);
	    (void) free(subalias);
	    (void) free(sup[i]);
	    continue;
	    }
	  aliases=(char **)vec_addlstlst((void **)aliases,(void *)sup[i]);
	  }
	(void) free(sup);
	}
      break;
    case 4	:	/*sorting alias list	*/
      (void) qsort((void *)aliases,vec_sizelstlst((void **)aliases),sizeof(char *),cmpali);
      break;
    case 5	:	/*duplicate no double	*/
      if (aliases!=(char **)0) { /*always	*/
	char **unique;

	unique=(char **)calloc(vec_sizelstlst((void **)aliases)+1,sizeof(char *));
	unique[0]=aliases[0];
	if (unique[0]!=(char *)0) {
	  int i,j;

	  for (i=1,j=0;aliases[i]!=(char *)0;i++) {
	    if (strcmp(unique[j],aliases[i])==0)
	      (void) free(aliases[i]);
	    else {
	      j++;
	      unique[j]=aliases[i];
	      }
	    }
	  }
        (void) free(aliases);
	aliases=unique;
	} 
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return aliases;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve data-base field	*/
/*	according one field name and email	*/
/*	address, field can be on more than on	*/
/*	table, search is done in a specific way.*/
/*						*/
/*	If field is not available a NULL ptr	*/
/*	is returned.				*/
/*						*/
/************************************************/
static char *getcommonfield(SQLFILE *sql,char *email,char *fieldname)

{
#define	PROBE0	"SELECT * FROM %s WHERE %s %s AND %s %s"
#define	PROBE1	"SELECT * FROM %s WHERE %s=%s"
#define	PROBE2	"SELECT * FROM %s WHERE %s %s"
#define	PROBE3	\
	"SELECT * FROM %s WHERE %s=(SELECT %s FROM %s WHERE domainname %s)"

char *field;
char *user;
char *cleanemail;
char *domain;
int phase;
int proceed;

field=(char *)0;
cleanemail=sql_cleanstr(sql,email);
user=getnsearch(sql,email);
domain=getdsearch(email);
phase=0;
proceed=true;
while (proceed==true) {
  char *probe;

  probe=(char *)0;
  switch (phase) {
    case 0	:	/*is it an aliases	*/
      (void) asprintf(&probe,PROBE0,ALIASES,"domain",domain,"username",user);
      break;
    case 1	:	/*is it in email table	*/
      (void) asprintf(&probe,PROBE1,USERS,"username",cleanemail);
      break;
    case 2	:	/*checking domain table	*/
      (void) asprintf(&probe,PROBE2,DOMS,"domainname",domain);
      break;
    case 3	:	/*checking group  table	*/
      (void) asprintf(&probe,PROBE3,CLMGRP,"groupid","groupid",DOMS,domain);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  if (probe!=(char *)0) {
    SQLRESULT *rs;
    
    if ((rs=sql_tupple(sql,probe))!=(SQLRESULT *)0) {
      if (sql_nbrtupple(sql,rs)>0) {
	char *ptr;

	if ((ptr=sql_getvalue(sql,rs,0,fieldname))!=(char *)0) {
          field=strdup(ptr);
	  proceed=false;	/*Found IT!		*/
	  }
	}
      (void) sql_dropresult(sql,rs);
      }
    (void) free(probe);
    }
  phase++;
  }
(void) free(domain);
(void) free(user);
(void) free(cleanemail);
return field;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve data-base value	*/
/*	according one field name and and email	*/
/*	address					*/
/*	to make sure the SPF protection apply	*/
/*	to him.					*/
/*						*/
/************************************************/
static int getcommonvalue(SQLFILE *sql,char *email,char *fieldname,int defval)

{
char *field;

if ((field=getcommonfield(sql,email,fieldname))!=(char *)0) {
  defval=atoi(field);
  (void) free(field);
  }
return defval;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return one override record.*/
/*	Record is ajusted to address the '*'	*/
/*	field contents.				*/
/*						*/
/************************************************/
static OVRTYP *getoneoverride(SQLFILE *sql,char *orgname,char *orgdomain,
					   SQLRESULT *rs,int num)

{
OVRTYP *ovr;
char *locname;
char *locdom;
int i;

ovr=(OVRTYP *)calloc(1,sizeof(OVRTYP));
locname=(char *)0;
locdom=(char *)0;
for (i=0;i<3;i++) {
  const char *dflt;
  const char *str;
  char **dst;

  dflt=(const char *)0;
  str=(const char *)0;
  dst=(char **)0;
  switch (i) {
    case 0	:
      str=sql_getvalue(sql,rs,num,"orgname");
      dflt=orgname;
      dst=&(locname);
      break;
    case 1	:
      str=sql_getvalue(sql,rs,num,"orgdom");
      dflt=orgdomain;
      dst=&(locdom);
      break;
    case 2	:
      str=sql_getvalue(sql,rs,num,"orgip");
      dflt="0/0";
      dst=&(ovr->orgip);
      break;
    }
  if (str==(char *)0)
    str="---";		/*NULL domain	*/
  if (strcmp(str,"*")==0) 
    str=dflt;
  *dst=sql_strdup(str);
  }
ovr->ovrid=(atol)(sql_getvalue(sql,rs,num,"ovrid"));
ovr->ovrride=(atoi)(sql_getvalue(sql,rs,num,"ovrride"));
(void) asprintf(&(ovr->orgemail),"%s@%s",locname,locdom);
(void) free(locdom);
(void) free(locname);
return ovr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if a table was	*/
/*	later than a specific date.		*/
/*						*/
/************************************************/
static time_t checkupdst(SQLFILE *sql,const char *table)

{
#define MAXREC	1	/*just need one record	*/
#define	SELUPD	"%s %s ORDER BY updated DESC LIMIT %d"

time_t update;
SQLRESULT *rs;

update=(time_t)0;
if ((rs=sql_tupple(sql,SELUPD,SELFROM,table,MAXREC))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)==MAXREC) {
    update=sql_tounixtime(sql,sql_getvalue(sql,rs,0,"updated"));
    }
  (void) sql_dropresult(sql,rs);
  }
return update;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update email logs		*/
/*						*/
/************************************************/
static int setlogs(SQLFILE *sql,DSPTYP *dsp,RCPTYP **rcpts)

{
#define	OPST	"gesdbc.c:setlogs,"

#define	LOGSFL1	"session,rmtip,rmthostname,originator,localip,taille,msgid"
#define	LOGSFL2	"subject,recipient,status,why,grpin,grpout,sender"
#define	INLOGS "INSERT INTO %s (%s,%s) VALUES (%s,%s,%s,%d,%s,%u,%u,%s)"
int done;

done=0;
if ((dsp!=(DSPTYP *)0)&&(rcpts!=(RCPTYP **)0)) {
  register char *subject;
  register char *hfrom;
  register char *msgid;
  register int grpin;
  char *str;
  int i;

  hfrom=sql_cleanstr(sql,dsp->hfrom);
  grpin=dsp->grpin;
  subject=sql_cleanstr(sql,dsp->subject);
  msgid=sql_cleanstr(sql,dsp->msgid);
  (void) asprintf(&str,"'%s','%s','%s',%s,'%s',%lu,%s",
		        dsp->session,dsp->peerip,dsp->peername,
			hfrom,dsp->locip,dsp->mtaille,msgid);
  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rloc;
    register char *rcpt;
    register char *inbehalf;
    register char *why;
    register char *locsub;
    register u_int grpout;

    rloc=rcpts[i];
    rcpt=sql_cleanstr(sql,rloc->rcpt);
    inbehalf=sql_cleanstr(sql,rloc->inbehalf);
    why=(char *)0;
    locsub=subject;
    grpout=rloc->grpid;
    switch (rloc->status) {
      case m_relayed	:
      case m_toforward	:
      case m_toresend	:
      case m_direct	:
        why=sql_cleanstr(sql,(char *)0);
	break;
      case m_rejected	:
      case m_noaddress	:
      case m_ipbanned	:
      case m_empty	:
      case m_discarded	:
      case m_virusfound	:
      case m_trouble	:
        why=sql_cleanstr(sql,rloc->why);
	locsub=why;
	break;
      case m_quarantine	:
        why=sql_cleanstr(sql,rloc->why);
	if (dsp->subject==(char *)0) 
	  locsub=why;
	break;
      default		:
	break;
      }
     if ((grpin==0)&&(grpout==0))
       (void) clm_alert(0,"%s session <%s>, sender=<%s> recipient=<%s> unassigned group! (rejected relaying?, bug?)",
			  OPST,dsp->session,rloc->inbehalf,rloc->rcpt);
    (void) sql_action(sql,INLOGS,EMLOGS,LOGSFL1,LOGSFL2,str,locsub,rcpt,rloc->status,why,grpin,grpout,inbehalf);
    why=clm_freestr(why);
    (void) free(inbehalf);
    (void) free(rcpt);
    done++;
    }
  (void) free(msgid);
  (void) free(subject);
  (void) free(hfrom);
  (void) free(str);
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get cron ptr within a cron	*/
/*	list.					*/
/*						*/
/************************************************/
char **getcronptr(char **crontab,char *cmd)

{
char **ptr;

ptr=(char **)0;
if (crontab!=(char **)0) {
  int i;

  for (i=0;crontab[i]!=(char *)0;i++) {
    if (strstr(crontab[i],cmd)!=crontab[i])
      continue;
    ptr=crontab+i;
    break;
    }
  }
return ptr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get cron entres for ONE	*/
/*						*/
/************************************************/
static char **getcron(SQLFILE *sql,int grpid,time_t curtime)

{
#define	SELCRON	"SELECT * FROM %s WHERE groupid=%d AND days LIKE '%%%s%%' ORDER BY cmd,groupid"
#define	UPDCRON	"UPDATE %s SET last='%s' WHERE cronid=%d"

static char *days[]={"Su","Mo","Tu","We","Th","Fr","Sa"};

char **crontab;
struct tm *tm;
SQLRESULT *rs;

crontab=(char **)0;
tm=localtime(&curtime);
if ((rs=sql_tupple(sql,SELCRON,CRON,grpid,days[tm->tm_wday]))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;

    for (i=0;i<nbr;i++) {
      char *cmd;
      time_t todotime;

      if (sscanf(sql_getvalue(sql,rs,i,"at"),"%d:%d:%d",&(tm->tm_hour),&(tm->tm_min),&(tm->tm_sec))!=3) {
	(void) clm_alert(0,"%s At <%s> format incorrect (DB trouble?)","gesdbc.c:getcron,",sql_getvalue(sql,rs,i,"at"));
	continue;
	}
      todotime=mktime(tm);
      if ((todotime<=curtime)&&(todotime>sql_tounixtime(sql,sql_getvalue(sql,rs,i,"last")))) {
	const char *last;

        last=sql_fromunixtime(sql,todotime);
        (void) sql_action(sql,UPDCRON,CRON,last,atoi(sql_getvalue(sql,rs,i,"cronid")));
        cmd=strdup(sql_getvalue(sql,rs,i,"cmd"));
        crontab=(char **)vec_addlstlst((void **)crontab,(void *)cmd);
        }
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return crontab;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to locat a banned remote	*/
/*	record					*/
/*						*/
/************************************************/
static SQLRESULT *locateban(SQLFILE *sql,char *ipnum,int *pos)

{
#define	OPLC	"gesdbc.c:locateban,"

#define	SELBAN	"SELECT * FROM %s WHERE ipbin>=%s AND ipbin<=%s ORDER BY ipnum DESC"

SQLRESULT *rs;
AFNTYP *curaf;
int foundit;
int phase;
int proceed;

*pos=-1;
rs=(SQLRESULT *)0;
curaf=afn_getipnum(ipnum);
foundit=false;
phase=0;
proceed=(curaf!=(AFNTYP *)0);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*computing limits	*/
      if (curaf!=(AFNTYP *)0) {	/*always	*/
	AFNTYP locafn;
	int mask;
	char *ipbin;
	char *limbas;
	char *limhaut;

	(void) memcpy(&locafn,curaf,sizeof(locafn));
	mask=16;
	if (locafn.afntype==AF_INET6)
  	  mask=48;
	(void) afn_maskip(&locafn,mask);
        ipbin=afn_ipnormal(&locafn);
	limbas=sql_toipbin(sql,ipbin);
	limhaut=sql_toipbin(sql,ipnum);
        if ((rs=sql_tupple(sql,SELBAN,BANTBL,limbas,limhaut))==(SQLRESULT *)0) {
          (void) clm_alert(0,"%s Unable to extract BAN information (bug?)",OPLC);
	  phase=999;	/*trouble trouble	*/
	  }
	(void) clm_freestr(limhaut);
	(void) clm_freestr(limbas);
	(void) clm_freestr(ipbin);
	}
      break;
    case 1	:	/*ssearchin data	*/
      if (rs!=(SQLRESULT *)0) {	/*always	*/
	int nbr;

        if ((nbr=sql_nbrtupple(sql,rs))>0) {
	  int i;

	  foundit=false;
	  for (i=0;i<nbr;i++) {
	    const char *normip;
	    char foundip[80];
	    AFNTYP *foundaf;

            normip=sql_getvalue(sql,rs,i,"ipnum");
            (void) strncpy(foundip,normip,sizeof(foundip)-1);
	    (void) afn_unformatip(foundip);
	    if ((foundaf=afn_getipnum(foundip))==(AFNTYP *)0) {
              (void) clm_alert(0,"%s Unable to convert ipnum <%s> (data-base contents bug?)",OPLC,foundip);
	      continue;
	      }
	    switch (afn_cmpipnum(curaf,foundaf,atoi(sql_getvalue(sql,rs,i,"mask")))) {
	      case true		:
	        *pos=i;
	        foundit=true;
		break;
	      case false	:
		break;
	      case -1		:
	        (void) clm_alert(0,"%s Unable to compare <%s> with <%s> (bug? errno=<%s>)",
				    OPLC,ipnum,foundip,strerror(errno));
		break;
	      }
	    foundaf=afn_freeipnum(foundaf);
	    if (foundit==true)
	      break;		/*we found a match	*/
	    }
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      if (foundit==false) {
        (void) sql_dropresult(sql,rs);
        rs=(SQLRESULT *)0;
        }
      proceed=false;
      break;
    }
  phase++;
  }
curaf=afn_freeipnum(curaf);
return rs;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update email transit record*/
/*						*/
/************************************************/
int dbc_settrtdata(SQLFILE *sql,QUETYP *queue)

{
#define	OPSE	"gesdbc.c:dbc_settrtdata,"

#define	UPDTRT	\
  "UPDATE %s SET why=%s,last=%s,next=%s,status=%d WHERE session='%s' AND mfrom=%s AND recipient=%s"
#define	INSTRT	"INSERT INTO %s VALUES ('%s',%s,%s,%s,%d,%s,%s,%s,%s,%d)"

static LCKTYP lck1={EMLTRT,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
char *last;
char *next;
char *why;
char *recipient;
char *receml;
char *orgeml;
int phase;
int proceed;

done=true;
(void) asprintf(&last,"'%s'",sql_fromunixtime(sql,queue->lasttime));
(void) asprintf(&next,"'%s'",sql_fromunixtime(sql,queue->nexttime));
(void) asprintf(&recipient,"%s@%s",queue->recname,queue->recdom);
why=sql_cleanstr(sql,queue->why);
receml=sql_cleanstr(sql,recipient);
orgeml=sql_cleanstr(sql,queue->orgeml);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*updating record	*/
      if (sql_action(sql,UPDTRT,EMLTRT,why,last,next,queue->status,
				       queue->session,orgeml,receml)==1)
        phase=999;	/*record already present*/
      break;
    case 1	:	/*locking database table*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPSE,EMLTRT);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*updating record again	*/
      if (sql_action(sql,UPDTRT,EMLTRT,why,last,next,queue->status,
				       queue->session,orgeml,receml)==1)
        phase++;	/*record just created! 	*/
      break;
    case 3	:	/*creating record	*/
      if (done==true) {	/*always		*/
	char *first;
	char *subject;

        (void) asprintf(&first,"'%s'",sql_fromunixtime(sql,queue->start));
        subject=sql_cleanstr(sql,queue->subject);
        if (sql_action(sql,INSTRT,EMLTRT,queue->session,first,last,next,queue->retry,
				         orgeml,receml,subject,why,queue->status)!=1) {
          (void) clm_alert(0,"%s Unable to create transaction table session=<%s>",
			      OPSE,queue->session);
          (void) sql_unlock_rollback(sql);
	  done=false;
	  phase=999;	/*trouble trouble	*/
	  }
       (void) free(subject);
       (void) free(first);
       }
      break;
    case 4	:	/*unlocking db table	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(orgeml);
(void) free(receml);
(void) free(why);
(void) free(recipient);
(void) free(next);
(void) free(last);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract an alias list if	*/
/*	there is one				*/
/*	available.				*/
/*						*/
/************************************************/
char **dbc_getaliaslist(SQLFILE *sql,char *rcpt)

{
char **known;
char **aliases;

known=(char **)0;
aliases=getaliaslist(sql,rcpt,&known,0);
known=(char **)vec_freelstlst((void **)known,VECFREE);
return aliases;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return user information if	*/
/*	available.				*/
/*						*/
/************************************************/
void dbc_getlocaluser(SQLFILE *sql,USRTYP *usr)

{
#define	OPGC	"gesdbc.c:dbc_getlocaluser,"

#define	SELUSR	"%s %s WHERE username=%s"

SQLRESULT *rs;
char *username;

usr->rcptflg=rcpt_unknown;
username=sql_cleanstr(sql,usr->username);
if ((rs=sql_tupple(sql,SELUSR,SELFROM,USERS,username))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) {
    PRFTYP *prf;
    GRPTYP *grp;

    prf=dbc_getprofile(sql,getcommonvalue(sql,usr->username,"prfid",1));
    usr->prfid=prf->prfid;
    usr->rcptflg=rcpt_islocal;
    usr->active=mngactive(usr->username,atoi(sql_getvalue(sql,rs,0,"active")));
    if ((grp=dbc_getgroup(sql,usr->username))!=(GRPTYP *)0) {
      usr->active&=grp->active;
      grp=dbc_freegroup(grp);
      }
    else
      (void) jrl_journalling(true,APJRL,"Unable to find goup definition "
					"for user <%s> (DB trouble?!)",usr->username);
    usr->upto=sql_tounixtime(sql,sql_getvalue(sql,rs,0,"upto"));
    if ((usr->upto>0)&&(usr->upto<time((time_t *)0))) 
      usr->active=false;
    if (usr->active==true) {
      unsigned long long usedbytes;
 
      /*we set user password if account is allowed to be accessed	*/
      usr->password=sql_strdup(sql_getvalue(sql,rs,0,"password"));
      /*checking if Email space reach its maximun value			*/
      usedbytes=(unsigned)atoll(sql_getvalue(sql,rs,0,"usedbytes"));
      if (prf->mxmgbytes<(unsigned int)(usedbytes/(1024*1024)))
	usr->rcptflg=rcpt_boxtb;
      }
    else {
      usr->password=strdup("!!");	/*user can't Be identified	*/
      usr->rcptflg=rcpt_inactive;
      }
    prf=dbc_freeprofile(prf);
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(username);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get the available remote	*/
/*	mentor list.				*/
/*						*/
/************************************************/
SRVLST **dbc_getsrvlst(SQLFILE *sql,char *srvname)

{
#define	GETMNT	"SELECT * FROM %s WHERE srvname='%s' ORDER BY preference LIMIT 10"

SRVLST **srvlst;
SQLRESULT *rs;

srvlst=(SRVLST **)0;
if ((rs=sql_tupple(sql,GETMNT,SRVTBL,srvname))!=(SQLRESULT *)0) {
  int nbr;
  int i;

  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    SRVLST *srvloc;

    srvloc=(SRVLST *)calloc(1,sizeof(SRVLST));
    srvloc->srvname=strdup(srvname);
    srvloc->srvhost=sql_strdup(sql_getvalue(sql,rs,i,"srvhost"));
    srvloc->srvport=sql_strdup(sql_getvalue(sql,rs,i,"srvport"));
    srvloc->srvaccess=sql_strdup(sql_getvalue(sql,rs,i,"srvaccess"));
    srvlst=(SRVLST **)vec_addlstlst((void **)srvlst,(void *)srvloc);
    }
  (void) sql_dropresult(sql,rs);
  }
return srvlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free the list of available	*/
/*	mentor.					*/
/*						*/
/************************************************/
SRVLST **dbc_freesrvlst(SRVLST **srvlst)

{
if (srvlst!=(SRVLST **)0) {
  int i;

  for (i=0;srvlst[i]!=(SRVLST *)0;i++) {
    (void) free(srvlst[i]->srvname);
    srvlst[i]->srvport=clm_freestr(srvlst[i]->srvport);
    srvlst[i]->srvhost=clm_freestr(srvlst[i]->srvhost);
    srvlst[i]->srvaccess=clm_freestr(srvlst[i]->srvaccess);
    (void) free(srvlst[i]);
    }
  (void) free(srvlst);
  srvlst=(SRVLST **)0;
  }
return srvlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to store the session result	*/
/*	inside the clement logs.		*/
/*						*/
/************************************************/
int dbc_setlogs(SQLFILE *sql,DSPTYP *dsp,RCPTYP **rcpts)

{
static LCKTYP lck1={EMLOGS,dbd_lockwrite};
static LCKTYP lck2={DOMS,dbd_lockread};
static LCKTYP lck3={DOMAL,dbd_lockread};
static LCKTYP *lcklst[]={&lck1,&lck2,&lck3,(LCKTYP *)0};

int done;
int phase;
int proceed;

done=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*locking data-base	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table","gesdbc.c:dbc_setlogs,",EMLOGS);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*updating logs		*/
      done=setlogs(sql,dsp,rcpts);	
      break;
    case 2	:	/*unlocking data base	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a log	*/
/*	record.					*/
/*						*/
/************************************************/
LOGTYP *dbc_freelogs(LOGTYP *mlog)

{
if (mlog!=(LOGTYP *)0) {
  mlog->subject=clm_freestr(mlog->subject);
  mlog->recipient=clm_freestr(mlog->recipient);
  mlog->hfrom=clm_freestr(mlog->hfrom);
  mlog->efrom=clm_freestr(mlog->efrom);
  mlog->rmtip=clm_freestr(mlog->rmtip);
  mlog->session=clm_freestr(mlog->session);
  (void) free(mlog);
  mlog=(LOGTYP *)0;
  }
return mlog;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a log 		*/
/*						*/
/************************************************/
LOGTYP *dbc_getlogs(SQLFILE *sql,char *session,char *recipient)

{
#define	OPLO	"gesdbc.c:dbc_getlogs,"

#define	SELLOG	"SELECT * FROM %s WHERE session='%s' AND recipient=%s"

LOGTYP *emlog;
char *rcpt;
SQLRESULT *rs;

emlog=(LOGTYP *)0;
rcpt=sql_cleanstr(sql,recipient);
if ((rs=sql_tupple(sql,SELLOG,EMLOGS,session,rcpt))!=(SQLRESULT *)0) {
  int nbr;

  nbr=sql_nbrtupple(sql,rs);
  if (nbr>0) {
    if (nbr>1)
      (void) clm_alert(0,"%s found %d %s records for session=<%s> and recipient<%s> (bug?)",
			  OPLO,nbr,EMLOGS,session,recipient);
    emlog=(LOGTYP *)calloc(1,sizeof(LOGTYP));
    emlog->session=strdup(session);
    emlog->rmtip=sql_strdup(sql_getvalue(sql,rs,0,"rmtip"));
    emlog->efrom=sql_strdup(sql_getvalue(sql,rs,0,"sender"));
    emlog->hfrom=sql_strdup(sql_getvalue(sql,rs,0,"originator"));
    emlog->subject=sql_strdup(sql_getvalue(sql,rs,0,"subject"));
    emlog->grpin=atoi(sql_strdup(sql_getvalue(sql,rs,0,"grpin")));
    emlog->recipient=strdup(recipient);
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(rcpt);
return emlog;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update session logs within	*/
/*	data-base.				*/
/*						*/
/************************************************/
int dbc_updlogs(SQLFILE *sql,char *session,char *recipient,MENTYP status)

{
#define	UPDLOGS "UPDATE %s set status=%d WHERE session='%s' AND recipient=%s"

int done;
char *rcpt;

done=false;
rcpt=sql_cleanstr(sql,recipient);
done=sql_action(sql,UPDLOGS,EMLOGS,status,session,rcpt);
(void) free(rcpt);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if binding info have	*/
/*	been updated?				*/
/*						*/
/************************************************/
time_t dbc_checkbind(SQLFILE *sql)

{
return checkupdst(sql,BINDING);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if relay info have	*/
/*	been updated?				*/
/*						*/
/************************************************/
time_t dbc_checkrelay(SQLFILE *sql)

{
return checkupdst(sql,RLYTBL);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get Binding information	*/
/*	request is done according a protocol	*/
/*	value.					*/
/*						*/
/************************************************/
BININF **dbc_getbinds(SQLFILE *sql,BININF **binds,PROTYP proto)

{
#define	OPGE	"gesdbc.c:dbc_getbinds,"
#define SELBIN	"SELECT * FROM %s WHERE active=1 AND 	\
		  bindid IN \
		 (SELECT bindid FROM %s WHERE groupid IN \
		 (SELECT groupid FROM %s WHERE active=1))"

static	const char *strproto[]={
		"smtp","smtps","pop3","mentor","Unknown",(char *)0
		};

SQLRESULT *rs;

if ((rs=sql_tupple(sql,SELBIN,BINDING,CLMBIND,CLMGRP))!=(SQLRESULT *)0) {
  const char *needproto;
  u_int nbr;
  u_int i;

  needproto=strproto[(int)proto];
  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    char *hostname;
    char *port;
    u_int iteration;
    char **list;

    if (strcmp(needproto,sql_getvalue(sql,rs,i,"protocol"))!=0)
      continue;
    iteration=(atoi)(sql_getvalue(sql,rs,i,"iteration"));
    port=sql_getvalue(sql,rs,i,"port");
    hostname=sql_getvalue(sql,rs,i,"binding");
    if ((list=dns_getiplist(hostname))==(char **)0) {
      (void) clm_alert(0,"%s Unable to get iplist for host <%s>",OPGE,hostname);
      }
    else {
      u_int j;

      for (j=0;list[j]!=(char *)0;j++) {
        BININF *entre;

        entre=dev_mkbindinf(proto,list[j],port,iteration);
        binds=(BININF **)vec_addlstlst((void **)binds,(void *)entre);
	}
      list=(char **)vec_freelstlst((void **)list,VECFREE);
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return binds;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract relaying data	*/
/*	included in data-base.			*/
/*						*/
/************************************************/
char **dbc_getrelay(SQLFILE *sql)

{
#define	OPRL	"gesdbc.c:dbc_getrelay,"
#define SELRLY	"SELECT * FROM %s"

char **strrly;
SQLRESULT *rs;

strrly=(char **)0;
if ((rs=sql_tupple(sql,SELRLY,RLYTBL))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;
    char *str;

    (void) asprintf(&str,"DB_table %s",RLYTBL);
    strrly=(char **)vec_addlstlst((void **)strrly,(void *)str);
    for (i=0;i<nbr;i++) {
      int groupid;
      char *ipnum;
      char *domain;
      char *upto;

      groupid=atoi(sql_getvalue(sql,rs,i,"groupid"));
      ipnum=sql_getvalue(sql,rs,i,"ipnumber");
      if ((domain=sql_getvalue(sql,rs,i,"domain"))==(char *)0)
	domain="-";
      if (sql_getvalue(sql,rs,i,"upto")!=(char *)0)
	upto=strdup(sql_getvalue(sql,rs,i,"upto"));
      else
	upto=strdup("");
      (void) asprintf(&str,"%d %s %s %s",groupid,ipnum,domain,upto);
      (void) free(upto);
      strrly=(char **)vec_addlstlst((void **)strrly,(void *)str);
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return strrly;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to dynamicaly add a domain	*/
/*	to group.				*/
/*						*/
/************************************************/
int dbc_adddomtogroup(SQLFILE *sql,GRPTYP *grp,char *domain)

{
#define	OPAD	"gesdbc.c:dbc_adddomtogroup,"

#define	UPDDOM	"UPDATE %s set lang=(SELECT lang FROM %s WHERE groupid=%d) WHERE domainname='%s'"
#define INDOM	"INSERT INTO %s (domainname,groupid,prfid,active) VALUES('%s',%d,%d,%d)"

static LCKTYP lck1={DOMS,dbd_lockwrite};
static LCKTYP lck2={CLMGRP,dbd_lockread};
static LCKTYP lck3={DOMAL,dbd_lockread};
static LCKTYP *lcklst[]={&lck1,&lck2,&lck3,(LCKTYP *)0};

int done;
char *action;
int phase;
int proceed;

done=true;
(void) asprintf(&action,UPDDOM,DOMS,CLMGRP,grp->groupid,domain);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*domain already added	*/
      if (sql_action(sql,action)>0)
	phase=999;	/*found at least one	*/
      break;
    case 1	:	/*locking the data-base	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPAD,DOMS);
	done=false;
	phase=999;
	}
      break;
    case 2	:	/*updating		*/
      if (sql_action(sql,action)>0) 
	phase++;	/*race condition 	*/
      break;
    case 3	:	/*creating		*/
      (void) free(action);
      (void) asprintf(&action,INDOM,DOMS,domain,grp->groupid,grp->prfid,(int)da_good);
      if (sql_action(sql,action)==0) {
	(void) clm_alert(0,"%s Unable to update %s table (bug?)",OPAD,DOMS);
        (void) sql_unlock_rollback(sql);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      else
	(void) sql_action(sql,UPDDOM,DOMS,CLMGRP,grp->groupid,domain);
      break;
    case 4	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(action);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a default domain	*/
/*	name. That domain is the oldest domain	*/
/*	defined within domain table in groupid 1*/
/*						*/
/************************************************/
char *dbc_getdfltdom(SQLFILE *sql)

{
#define	GETDFLD	"SELECT * FROM %s WHERE groupid=1 AND active=1 ORDER BY first ASC"
char *dfltdom;
SQLRESULT *rs;

dfltdom=(char *)0;
if ((rs=sql_tupple(sql,GETDFLD,DOMS))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) 
    dfltdom=sql_strdup(sql_getvalue(sql,rs,0,"domainname"));
  (void) sql_dropresult(sql,rs);
  }
return dfltdom;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a default originator*/
/*	according domain provided.		*/
/*	For now the default from is always the	*/
/*	same value.				*/
/*						*/
/************************************************/
char *dbc_getdfltfrom(SQLFILE *sql,char *domain)

{
char *dfltfrom;
char *dfltdom;

if ((dfltdom=dbc_getdfltdom(sql))==(char *)0)
  dfltdom=domain;
(void) asprintf(&dfltfrom,"%s@%s",POSTM,dfltdom);
(void) free(dfltdom);
return dfltfrom;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a local*/
/*	user definition.			*/
/*						*/
/************************************************/
USRTYP *dbc_freelocaluser(USRTYP *localusr)

{
if (localusr!=(USRTYP *)0) {
  localusr->password=clm_freestr(localusr->password);
  localusr->username=clm_freestr(localusr->username);
  (void) free(localusr);
  localusr=(USRTYP *)0;
  }
return localusr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by group	*/
/*	definition structure.			*/
/*						*/
/************************************************/
GRPTYP *dbc_freegroup(GRPTYP *group)

{
if (group!=(GRPTYP *)0) {
  group->timezone=clm_freestr(group->timezone);
  group->exitip=clm_freestr(group->exitip);
  group->ldapbase=clm_freestr(group->ldapbase);
  group->authport=clm_freestr(group->authport);
  group->authrmt=clm_freestr(group->authrmt);
  group->smtpport=clm_freestr(group->smtpport);
  group->smtprmt=clm_freestr(group->smtprmt);
  group->grpname=clm_freestr(group->grpname);
  (void) free(group);
  group=(GRPTYP *)0;
  }
return group;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a group*/
/*	list.					*/
/*						*/
/************************************************/
GRPTYP **dbc_freegrouplst(GRPTYP **groups)

{
if (groups!=(GRPTYP **)0) {	/*always*/
  int i;

  for (i=0;groups[i]!=(GRPTYP *)0;i++) 
    groups[i]=dbc_freegroup(groups[i]);
  (void) free(groups);
  groups=(GRPTYP **)0;
  }
return groups;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the group id related*/
/*	to a specific email.			*/
/*						*/
/************************************************/
u_int dbc_getgroupid(SQLFILE *sql,char *rcpt)

{
return getgroupid(sql,rcpt);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve group record	*/
/*	related to a specific group ID.		*/
/*						*/
/************************************************/
GRPTYP *dbc_getgroupbyid(SQLFILE *sql,int groupid)

{
#define	SELCLM	"%s %s WHERE groupid=%d"

GRPTYP *grp;
SQLRESULT *rs;

grp=(GRPTYP *)0;
if ((rs=sql_tupple(sql,SELCLM,SELFROM,CLMGRP,groupid))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) {
    char *smtprmt;
    char *authrmt;

    grp=(GRPTYP *)calloc(1,sizeof(GRPTYP));
    grp->groupid=groupid;
    grp->active|=(atoi)(sql_getvalue(sql,rs,0,"active"));
    grp->prfid=(atoi)(sql_getvalue(sql,rs,0,"prfid"));
    grp->grpname=sql_strdup(sql_getvalue(sql,rs,0,"groupname"));
    grp->upto=getuptograce(sql,rs,0);
    grp->smtpport=strdup(SMTPPORT);
    grp->authprot=pro_ukn;
    smtprmt=sql_getvalue(sql,rs,0,"remotesmtp");
    if (smtprmt!=(char *)0) {
      char *strport;
      char *exitip;

      grp->smtprmt=strdup(smtprmt);
      if ((strport=sql_getvalue(sql,rs,0,"remoteport"))!=(char *)0) {
	if (atoi(strport)!=0) {
	  (void) free(grp->smtpport);
          grp->smtpport=strdup(strport);
	  }
	}
      if ((exitip=sql_getvalue(sql,rs,0,"exitip"))==(char *)0)
	exitip=ANYADDR;
      grp->exitip=strdup(exitip);
      }
    grp->timezone=sql_strdup(sql_getvalue(sql,rs,0,"timezone"));
    grp->authport=strdup(POP3SPORT);
    grp->authprot=pro_pop3s;
    authrmt=sql_getvalue(sql,rs,0,"authserver");
    if (authrmt!=(char *)0) {
      char *authport;
      char *authprot;

      grp->authrmt=strdup(authrmt);
      if ((authport=sql_getvalue(sql,rs,0,"authport"))!=(char *)0) {
	if (atoi(authport)!=0) {
	  (void) free(grp->authport);
	  grp->authport=strdup(authport);
	  }
	}
      if ((authprot=sql_getvalue(sql,rs,0,"authproto"))!=(char *)0) {
	grp->authprot=dev_getproto(authprot);
	switch (grp->authprot) {
	  case pro_ldap	:
	  case pro_ldaps:
	    grp->ldapbase=strdup(sql_getvalue(sql,rs,0,"ldapbase"));
	    break;
          default	:
	    break;
	  }
	}
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return grp;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract group information	*/
/*	from data-base.				*/
/*	If we are not able to retreive a group	*/
/*	from the domain, we return a NUL pointer*/
/*						*/
/************************************************/
GRPTYP *dbc_getgroup(SQLFILE *sql,char *rcpt)

{
GRPTYP *grp;
u_int groupid;

grp=(GRPTYP *)0;
groupid=0;
if ((groupid=getgroupid(sql,rcpt))!=0) {
  if ((grp=dbc_getgroupbyid(sql,groupid))==(GRPTYP *)0)
    (void) jrl_journalling(true,APJRL,"Unable to find goup definition for <%s> (DB trouble?!)",rcpt);
  }
return grp;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract group information	*/
/*	list from data-base.			*/
/*						*/
/************************************************/
GRPTYP **dbc_getgrouplst(SQLFILE *sql)

{
#define	SELGRP	"SELECT * FROM %s ORDER BY groupid ASC"

GRPTYP **grplst;
SQLRESULT *rs;

grplst=(GRPTYP **)0;
if ((rs=sql_tupple(sql,SELGRP,CLMGRP))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;

    for (i=0;i<nbr;i++) {
      int groupid;
      GRPTYP *grp;

      groupid=(atoi)(sql_getvalue(sql,rs,i,"groupid"));
      if ((grp=dbc_getgroupbyid(sql,groupid))!=(GRPTYP *)0) 
        grplst=(GRPTYP **)vec_addlstlst((void **)grplst,(void *)grp);
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return grplst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract the default group	*/
/*	in case we are unable to locate a group	*/
/*	from the recipient or orignator.	*/
/*						*/
/************************************************/
char *dbc_getdfltgrpname(SQLFILE *sql)

{
char *dfltgrpname;
char *dfltdom;

dfltgrpname=(char *)0;
if ((dfltdom=dbc_getdfltdom(sql))!=(char *)0) {
  GRPTYP *dfltgrp;

  if ((dfltgrp=dbc_getgroup(sql,dfltdom))!=(GRPTYP *)0) {
    dfltgrpname=strdup(dfltgrp->grpname);
    dfltgrp=dbc_freegroup(dfltgrp);
    }
  (void) free(dfltdom);
  }
return dfltgrpname;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the BLSlists from	*/
/*	the data-base within an allocated memory*/
/*						*/
/************************************************/
char **dbc_getblssites(SQLFILE *sql,int blsid,int af_net)

{
#define	GETBLS	"SELECT * FROM %s WHERE blsid=%d"
char **sblsites;
SQLRESULT *rs;

sblsites=(char **)0;
if ((rs=sql_tupple(sql,GETBLS,BLSHST,blsid))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;
 
    for (i=0;i<nbr;i++) {
      char *ipmode;

      ipmode=(char *)0;
      switch (af_net) {
	case 0		:	/*we take all BLS site	*/
	  ipmode="ipv";
	  break;
	case AF_INET	:	/*Only IPV4 Site	*/
	  ipmode="ipv4";
	  break;
	case AF_INET6	:	/*Only IPV6 Site	*/
	  ipmode="ipv6";
	  break;
	default		:
	  (void) clm_alert(0,"%s Unexpected af_net value '%d' (bug?)","gesdbc.c:dbc_getblssites",af_net);
	  break;
	}
      if (strstr(sql_getvalue(sql,rs,i,"ipmode"),ipmode)!=(char *)0) {
        char *ptr;

        ptr=strdup(sql_getvalue(sql,rs,i,"blsserver"));
        sblsites=(char **)vec_addlstlst((void **)sblsites,(void *)ptr);
	}
      }
    (void) sql_dropresult(sql,rs);
    }
  }
return sblsites;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to prepare a user info struct	*/
/*						*/
/************************************************/
USRTYP *dbc_mkuserinfo(char *rcpt,RCPTFLG status)

{
USRTYP *usr;

usr=(USRTYP *)calloc(1,sizeof(USRTYP));
usr->username=strdup(rcpt);
usr->rcptflg=status;
return usr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if a specific user	*/
/*	is a local one.				*/
/*						*/
/************************************************/
void dbc_getuserinfo(SQLFILE *sql,USRTYP *usr)

{
int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*active domain?	*/
      if (dbc_getactive(sql,eml_getdomain(usr->username))!=1)
	phase=999;	/*domain NOT active	*/
      break;
    case 1	:	/*is user localy known	*/
      (void) dbc_getlocaluser(sql,usr);
      break;
    case 2	:	/*is user an alias	*/
      if ((usr->aliases=dbc_getaliaslist(sql,usr->username))!=(char **)0) {
	if ((usr->numalias=vec_sizelstlst((void **)usr->aliases))>0) {
	  if (usr->rcptflg==rcpt_unknown)
	    usr->rcptflg=rcpt_isalias;
	  else
	    usr->rcptflg=rcpt_islocalias;
	  }
	phase=999;	/*its an alias		*/
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by	*/
/*	user information			*/
/*						*/
/************************************************/
USRTYP *dbc_freeuserinfo(USRTYP *user)

{
if (user!=(USRTYP *)0) {
  user->aliases=(char **)vec_freelstlst((void **)user->aliases,VECFREE);
  user->password=clm_freestr(user->password);
  user->username=clm_freestr(user->username);
  (void) free(user);
  user=(USRTYP *)0;
  }
return user;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to build a structure to check	*/
/*	recipient name against a profile	*/
/*						*/
/************************************************/
PRFTYP *dbc_getprofile(SQLFILE *sql,int prfid)

{
#define	GETPRF	"SELECT * FROM %s WHERE prfid=%d"

PRFTYP *profile;
SQLRESULT *rs;

profile=(PRFTYP *)0;
if ((rs=sql_tupple(sql,GETPRF,USEPRF,prfid))!=(SQLRESULT *)0) {
  profile=(PRFTYP *)calloc(1,sizeof(PRFTYP));
  /*lets set minimal values				*/
  profile->prfid=1;		/*external account	*/
  profile->emailsz=10*1024;	/*ten K max		*/
  profile->checkrmtspf=1;
  profile->checkrmtmx=1;
  profile->bulk=3;		/*very minimal value	*/
  profile->lang=DFLTLNG;
  profile->spamlim=NAN;
  if (sql_nbrtupple(sql,rs)>0) {
    const char *ptr;

    profile->prfid=prfid;
    profile->system=atoi(sql_getvalue(sql,rs,0,"system"));
    profile->emailsz=atol(sql_getvalue(sql,rs,0,"emailsize"));
    profile->checkrmtspf=atoi(sql_getvalue(sql,rs,0,"checkrmtspf"));
    profile->checkrmtmx=atoi(sql_getvalue(sql,rs,0,"checkrmtmx"));
    profile->needreverse=atoi(sql_getvalue(sql,rs,0,"needreverse"));
    profile->mxmgbytes=atol(sql_getvalue(sql,rs,0,"mxmgbytes"));
    profile->bulk=atoi(sql_getvalue(sql,rs,0,"bulk"));
    profile->blsid=atoi(sql_getvalue(sql,rs,0,"blsid"));
    profile->maxperday=atoi(sql_getvalue(sql,rs,0,"maxperday"));
    ptr=sql_getvalue(sql,rs,0,"lang");
    if (ptr!=(char *)0)
      profile->lang=atoi(ptr);
    ptr=sql_getvalue(sql,rs,0,"spamlim");
    if (ptr!=(char *)0)
      profile->spamlim=atof(ptr);
    }
  (void) sql_dropresult(sql,rs);
  }
return profile;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	profile definition			*/
/*						*/
/************************************************/
PRFTYP *dbc_freeprofile(PRFTYP *profile)

{
if (profile!=(PRFTYP *)0) {
  (void) free(profile);
  profile=(PRFTYP *)0;
  }
return profile;
}
/*
^L
*/
/************************************************/
/*	Procedure to return a profil, if a	*/
/*	specific profil is not found, send the	*/
/*	originator profile (profile 0)		*/
/*						*/
/************************************************/
PRFTYP *dbc_getemailprofile(SQLFILE *sql,char *user)

{
PRFTYP *prf;

prf=dbc_getprofile(sql,getcommonvalue(sql,user,"prfid",1));
prf->active=mngactive(user,getcommonvalue(sql,user,"active",0));
return prf;
}
/*
^L
*/
/************************************************/
/*	Procedure to free memory used by an	*/
/*	override structure.			*/
/*						*/
/************************************************/
OVRTYP **dbc_freeoverride(OVRTYP **ovrlst)

{
if (ovrlst!=(OVRTYP **)0) {
  int i;

  for (i=0;ovrlst[i]!=(OVRTYP *)0;i++) {
    register OVRTYP *ptr;

    ptr=ovrlst[i];
    ptr->orgip=clm_freestr(ptr->orgip);
    ptr->orgemail=clm_freestr(ptr->orgemail);
    (void) free(ptr);
    }
  (void) free(ovrlst);
  ovrlst=(OVRTYP **)0;
  }
return ovrlst;
}
/*
^L
*/
/************************************************/
/*	Procedure to insert an overrid record	*/
/*	within data-base.			*/
/*						*/
/************************************************/
int dbc_setoverride(SQLFILE *sql,char *recipient,char *originator,char *orgdom,char *iporg,int cidr,OVRENUM ovrride)

{
#define	OPOV	"gesdbc.c:dbc_setoverride,"

#define	INSOVR1	"INSERT INTO %s (ovrride,orgname,orgdom,orgip,recname,recdom) "
#define	INSOVR2	"VALUES (%d,%s,'%s','%s',%s,'%s')"
#define	UPDOVR1	"UPDATE %s SET ovrride=%d,lasthit=NOW() WHERE "
#define	UPDOVR2	"recdom='%s' AND recname=%s AND orgip='%s' AND orgdom='%s' and orgname=%s"
		 
static LCKTYP lck1={OVRRIDE,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
char *recname;
char *orgname;
char *recdomain;
char *ipclass;
char *action;
int phase;
int proceed;

done=true;
recdomain=eml_splitdomain(recipient);
recname=sql_cleanstr(sql,recipient);	/*MUST BE after splitdomain	*/
orgname=sql_cleanstr(sql,originator);
ipclass=dns_getipclass(iporg,cidr);
(void) asprintf(&action,UPDOVR1 UPDOVR2,OVRRIDE,ovrride,recdomain,recname,ipclass,orgdom,orgname);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*lets update the record*/
      if (sql_action(sql,action)>0)
	phase=999;	/*found at least one	*/
      break;
    case 1	:	/*setting lock		*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPOV,OVRRIDE);
	done=false;
	phase=999;
	}
      break;
    case 2	:	/*updating		*/
      if (sql_action(sql,action)>0) 
	phase++;	/*race condition 	*/
      break;
    case 3	:	/*creating		*/
      (void) free(action);
      (void) asprintf(&action,INSOVR1 INSOVR2,OVRRIDE,ovrride,orgname,orgdom,ipclass,recname,recdomain);
      if (sql_action(sql,action)==0) {
	(void) clm_alert(0,"%s Unable to update %s table (bug?)",OPOV,OVRRIDE);
        (void) sql_unlock_rollback(sql);
	done=false;
	phase=999;	/*trouble trouble	*/	
	}
      break;
    case 4	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    } 
  phase++;
  }
(void) free(action);
(void) free(ipclass);
(void) free(orgname);
(void) free(recname);
return done;
}
/*
^L
*/
/************************************************/
/*	Procedure to return an override level	*/
/*	according special override directive	*/
/*	found with the data-base.		*/
/*						*/
/************************************************/
OVRTYP **dbc_getoverride(SQLFILE *sql,char *from,char *to)

{
#define	GETOVR	"SELECT * FROM %s WHERE recdom %s AND recname %s"

OVRTYP **ovrlst;
register const char *cd0;
char *orgdomain;
char *orgname;
char *recdomain;
char *recname;
int phase;
int proceed;

ovrlst=(OVRTYP **)0;
cd0="= '*'";
orgname=strdup(from);
orgdomain=eml_splitdomain(orgname);
recname=getnsearch(sql,to);
recdomain=getdsearch(to);
phase=0;
proceed=true;
while (proceed==true) {
  register const char *cd1;
  register const char *cd2;

  cd1=(char *)0;
  cd2=(char *)0;
  switch (phase) {
    case 0	:	/*overide for this user	*/
      cd1=recdomain;
      cd2=recname;
      break;
    case 1	:	/*overiding this username*/
      cd1=cd0;
      cd2=recname;
      break;
    case 2	:	/*overiding this domain	*/
      cd1=recdomain;
      cd2=cd0;
      break;
    case 3	:	/*overiding everything?	*/
      cd1=cd0;
      cd2=cd0;
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  if ((cd1!=(char *)0)&&(cd2!=(char *)0)) {
    SQLRESULT *rs;

    if ((rs=sql_tupple(sql,GETOVR,OVRRIDE,cd1,cd2))!=(SQLRESULT *)0) {
      int nbr;

      if ((nbr=sql_nbrtupple(sql,rs))>0) {
        int i;
 
        for (i=0;i<nbr;i++) {
	  register OVRTYP *ptr;
	
	  ptr=getoneoverride(sql,orgname,orgdomain,rs,i);
	  ovrlst=(OVRTYP **)vec_addlstlst((void **)ovrlst,(void *)ptr);
	  }
	}
      (void) sql_dropresult(sql,rs);
      }
    }
  phase++;
  }
(void) free(recdomain);
(void) free(recname);
(void) free(orgname);
return ovrlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update the override stamp	*/
/*	to mark the last hit.			*/
/*						*/
/************************************************/
void dbc_overhit(SQLFILE *sql,OVRTYP *over)

{
#define	OPOH	"gesdbc.c:dbc_overhit,"

#define	UPDOVR	"UPDATE %s SET lasthit=NOW() WHERE ovrid=%lu"

if (sql_action(sql,UPDOVR,OVRRIDE,over->ovrid)==0)
  (void) clm_alert(0,"%s Unable to update %s table (ovrid=%d)",OPOH,OVRRIDE,over->ovrid);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a list	*/
/*	file in quarantine.			*/
/*						*/
/************************************************/
EXTTYP **dbc_freeextract(EXTTYP **exts)

{
if (exts!=(EXTTYP **)0) {
  int i;

  for (i=0;exts[i]!=(EXTTYP *)0;i++) {
    (void) free(exts[i]->recipient); 
    (void) free(exts[i]->session); 
    (void) free(exts[i]);
    }
  (void) free(exts);
  exts=(EXTTYP **)0;
  }
return exts;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract from the database	*/
/*	a list of session/recipient used to	*/
/*	extract quarantined email from		*/
/*	depository.				*/
/*						*/
/************************************************/
EXTTYP **dbc_getextract(SQLFILE *sql)

{
#define	OPGX	"gesdbc.c:dbc_getextract,"

#define	DELEXT	"DELETE FROM %s WHERE session='%s' AND recipient=%s"
#define	SELEXT	"SELECT * FROM %s ORDER BY stamp ASC"

static LCKTYP lck1={EXTRACT,dbd_lockwrite};
static LCKTYP lck2={EMLOGS,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,&lck2,(LCKTYP *)0};

EXTTYP **exts;
SQLRESULT *rs;
int nbr;
int phase;
int proceed;

exts=(EXTTYP **)0;
rs=(SQLRESULT *)0;
nbr=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting lock access	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPGX,EXTRACT);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*extracting record	*/
      if ((rs=sql_tupple(sql,SELEXT,EXTRACT))!=(SQLRESULT *)0){
	nbr=sql_nbrtupple(sql,rs);
	if (nbr>0) {
	  int i;

	  for (i=0;i<nbr;i++) {
	    char *session;
	    char *recipient;
	    char *rcpt;
	    MENTYP status;
	    int trouble;

	    trouble=false;
	    session=sql_getvalue(sql,rs,i,"session");
	    recipient=sql_getvalue(sql,rs,i,"recipient");
	    rcpt=sql_cleanstr(sql,recipient);
	    status=getemlogstatus(sql,session,rcpt);
	    if (status!=m_undefined) {
	      EXTTYP *extloc;

	      extloc=(EXTTYP *)calloc(1,sizeof(EXTTYP));
	      extloc->session=sql_strdup(session);
	      extloc->recipient=sql_strdup(recipient);
	      extloc->status=status;
	      exts=(EXTTYP **)vec_addlstlst((void **)exts,(void *)extloc);
      	      if (sql_action(sql,DELEXT,EXTRACT,session,rcpt)==0) {
        	(void) clm_alert(0,"%s Unable to delete record from table %s (session='%s',recipient='%s' bug?)",
				    OPGX,EXTRACT,session,recipient);
		trouble=true;
		}
	      if (dbc_updlogs(sql,session,recipient,m_progress)==false) {
        	(void) clm_alert(0,"%s Unable to update record from table %s (session='%s',recipient='%s' bug?)",
				    OPGX,EMLOGS,session,recipient);
		trouble=true;
		}
	      if (trouble==true) {
      		(void) sql_unlock_rollback(sql);
		phase=999;	/*trouble trouble*/
		break;
		}
	      }
	    (void) free(rcpt);
	    }
	  }
        (void) sql_dropresult(sql,rs);
	}
      break;
    case 2	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return exts;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to insert a volatil entry	*/
/*	about an internal user within the data	*/
/*	base. Purpose is to avoid to request	*/
/*	over and over if user is existing	*/
/*	to the internal server.			*/
/*						*/
/************************************************/
int dbc_setintusr(SQLFILE *sql,char *rcpt,RCPTFLG rcptflg)

{
#define	OPIN	"gesdbc.c:dbc_setintusr"

#define	UPDIUSR	"UPDATE %s SET status=%d,upto='%s',lastchk=NOW() WHERE username=%s"
#define	INSIUSR	"INSERT INTO %s (username,status,upto,lastchk) VALUES(%s,%d,'%s',NOW())"

static LCKTYP lck1={INTUSR,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
register const char *UPTO;
register char *cleanrcpt;
long offset;
int phase;
int proceed;

done=true;
UPTO="";
cleanrcpt=sql_cleanstr(sql,rcpt);
offset=31*H24;	/*1 Month			*/
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*updating record	*/
      switch (rcptflg) {
	case rcpt_notchecked	:
	  offset=60;		/*One minute	*/
	  break;
	case rcpt_unknown	:
	  offset=(30*60);	/*30 minutes	*/
	  break;
	case rcpt_forward	:
	default			:
	  break;
	}
      UPTO=sql_fromunixtime(sql,time((time_t *)0)+offset);
      if (sql_action(sql,UPDIUSR,INTUSR,rcptflg,UPTO,cleanrcpt)==1)
        phase=999;	/*record already present*/
      break;
    case 1	:	/*locking database table*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPIN,INTUSR);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*double checking record*/
      if (sql_action(sql,UPDIUSR,INTUSR,rcptflg,UPTO,cleanrcpt)==1) 
        phase++;	/*record just created! 	*/
      break;
    case 3	:	/*creating record	*/
      if (sql_action(sql,INSIUSR,INTUSR,cleanrcpt,rcptflg,UPTO)!=1) {
        (void) clm_alert(0,"%s Unable to create internal user =<%s>",OPIN,rcpt);
        (void) sql_unlock_rollback(sql);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*unlocking db table	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(cleanrcpt);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if we have already	*/
/*	an internal user status within the	*/
/*	database.				*/
/*						*/
/************************************************/
RCPTFLG dbc_getintusr(SQLFILE *sql,char *rcpt)

{
#define	GETINTUSR	"SELECT * FROM %s WHERE username=%s"

#define	CHKDELAY	(30*60)	/*minimal check delay	*/

RCPTFLG rcptflg;
register char *cleanrcpt;
SQLRESULT *rs;

rcptflg=rcpt_uknstatus;
cleanrcpt=sql_cleanstr(sql,rcpt);
if ((rs=sql_tupple(sql,GETINTUSR,INTUSR,cleanrcpt))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    time_t lastchk;

    lastchk=CHKDELAY;	/*check every 30 minutes	*/
    rcptflg=(RCPTFLG)atoi(sql_getvalue(sql,rs,0,"status"));
    switch (rcptflg) {
      case rcpt_unknown		:
      case rcpt_forward		:
	lastchk*=6;	/*double checking every 12 hours	*/
			/*NO BREAK			*/
      case rcpt_notchecked	:
        lastchk+=sql_tounixtime(sql,sql_getvalue(sql,rs,0,"lastchk"));
        if (lastchk<time((time_t *)0))
	  rcptflg=rcpt_uknstatus;	/*asking for a refresh	*/
	break;
      default		:
	break;
      }
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(cleanrcpt);
return rcptflg;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to cleanup data-base dynamique*/
/*	table.					*/
/*	If the 'flush' variable is set, delete	*/
/*	all table contents, regardless of the	*/
/*	current time.				*/
/*						*/
/************************************************/
void dbc_dbcleanup(SQLFILE *sql)

{
#define	DELINU	"DELETE FROM %s WHERE upto<NOW()"
#define DELTRB	"DELETE FROM %s WHERE upto<'%s' AND progress=%d"

int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:
      break;
    case 2	:	/*cleaning old intusers	*/
      (void) sql_action(sql,DELINU,INTUSR);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to create a set of queued	*/
/*	entry about Email to be transited.	*/
/*						*/
/************************************************/
int dbc_createqueues(SQLFILE *sql,QUETYP **queues)

{
#define	OPCR	"gesdbc.c:dbc_createqueue,"

int numdone;
int proceed;
int phase;

numdone=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*locking database	*/
      if (sql_lock_action(sql)<0) {
        (void) clm_alert(0,"%s Unable to dbd_locksoft %s table",OPCR,INQUEUE);
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*inserting record	*/
      if (queues!=(QUETYP **)0) {
	int i;

	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  if (setqueue(sql,queues[i])==false) {
	    numdone=0;
	    (void) sql_unlock_rollback(sql);
	    phase=999;	/*trouble trouble	*/
	    break;
	    }
	  numdone++;
	  }
        }
      break;
    case 2	:	/*committing result	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return numdone;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to collect old inqueue 	*/
/*	transaction				*/
/*						*/
/************************************************/
QUETYP **dbc_getoldqueues(SQLFILE *sql,time_t curtime,int progress)

{
#define SELQDEL	"SELECT * FROM %s WHERE start<'%s' AND progress=%d"

QUETYP **queues;
SQLRESULT *rs;
const char *strcurt;

queues=(QUETYP **)0;
strcurt=sql_fromunixtime(sql,curtime);
if ((rs=sql_tupple(sql,SELQDEL,INQUEUE,strcurt,progress))!=(SQLRESULT *)0) {
  int nbr;
  int i;

  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    QUETYP *q;

    q=getonequeue(sql,rs,i);
    queues=(QUETYP **)vec_addlstlst((void **)queues,(void *)q);
    }
  (void) sql_dropresult(sql,rs);
  }
return queues;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to collect Email ready to be	*/
/*	transmitted. The Flag 'progress' is	*/
/*	set to 1 (inprogress).			*/
/*						*/
/************************************************/
QUETYP **dbc_getqueues(SQLFILE *sql,int max)

{
#define	OPGQ	"gesdbc.c:dbc_getqueues,"

#define	INPROG	"UPDATE %s SET progress=%d,last=NOW() WHERE session='%s' AND transit=%d"

QUETYP **queues;
int proceed;
int phase;

queues=(QUETYP **)0;
proceed=true;
phase=0;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*locking EMLOG READ	*/
      if (sql_lock_action(sql)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPGQ,INQUEUE);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*extracting ready	*/
      if ((queues=getpendingqueue(sql,max))!=(QUETYP **)0) {
	int i;

	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  if (sql_action(sql,INPROG,INQUEUE,QPROG,queues[i]->session,queues[i]->transit)!=1) {
            (void) clm_alert(0,"%s Unable flag queue 'progress' session=<%s> transit='%d' (bug?)",
				OPCR,queues[i]->session,queues[i]->transit);
            (void) sql_unlock_rollback(sql);
	    queues=que_freequeues(queues);
	    phase=999;	/*trouble trouble	*/
	    break;
	    }
	  }
	}
      break;
    case 2	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return queues;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update inqueue email data	*/
/*	transmitted. The Flag 'progress' is	*/
/*	set to 1 (inprogress).			*/
/*						*/
/************************************************/
int dbc_updqueue(SQLFILE *sql,QUETYP *q)

{
#define	UPDQUE	\
 "UPDATE %s SET next='%s',last='%s',upto='%s',retry='%d',progress=%d,status=%d,why=%s,traces=%s,numtrace=%lu WHERE session='%s' AND transit=%d"

int phase;
int proceed;
int num;
char *cleanlogs;
char *last;
char *next;
char *upto;
char *why;

phase=0;
proceed=true;
cleanlogs=(char *)0;
last=strdup(sql_fromunixtime(sql,q->lasttime));
next=strdup(sql_fromunixtime(sql,q->nexttime));
upto=strdup(sql_fromunixtime(sql,q->upto));
why=sql_cleanstr(sql,q->why);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*default logs		*/
      if (q->logs!=(char **)0) {
        int i;
	char *logs;

	logs=(char *)0;
	for (i=0;q->logs[i]!=(char *)0;i++) {
	  char *sublog;

	  (void) asprintf(&sublog,"%s\n",q->logs[i]);
	  logs=clm_astrcat(logs,sublog);
	  (void) free(sublog);
	  }
	cleanlogs=sql_cleanstr(sql,logs);
	(void) free(logs);
	phase++;	/*q->logs scanned	*/
        }
      break;
    case 1	:	/*empty logs		*/
      cleanlogs=strdup("''");
      break;
    default	:
      num=sql_action(sql,UPDQUE,INQUEUE,next,last,upto,q->retry,q->progress,q->status,why,cleanlogs,q->numtrace,
				q->session,q->transit);
      (void) free(cleanlogs);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(why);
(void) free(upto);
(void) free(next);
(void) free(last);
return num;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to move an entrie from the	*/
/*	inqueue list in ready mode.		*/
/*						*/
/************************************************/
int dbc_rdyqueue(SQLFILE *sql,QUETYP *queue)

{
#define	RDYQUE	"UPDATE %s SET progress=0 WHERE session='%s' AND transit=%d"

int done;

done=true;
if (queue!=(QUETYP *)0) {
  if (sql_action(sql,RDYQUE,INQUEUE,queue->session,queue->transit)<1) {
    (void) clm_alert(0,"%s Unable to make inqueue session=<%s> transit='%d' ready",
			"gesdbc.c:dbc_rdyqueue,",queue->session,queue->transit);
    done=false;
    }
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove an entrie from the	*/
/*	inqueue list.				*/
/*						*/
/************************************************/
int dbc_delqueue(SQLFILE *sql,QUETYP *queue)

{
#define	FLGQUE	"UPDATE %s SET progress=%d WHERE session='%s' AND transit=%d"
#define	DELQUE	"DELETE FROM %s WHERE session='%s' AND transit=%d"

int done;

done=true;
if (queue!=(QUETYP *)0) {
  int status;

  if (debug>0)
    status=sql_action(sql,FLGQUE,INQUEUE,QDEL,queue->session,queue->transit);
  else {
    status=sql_action(sql,DELQUE,INQUEUE,queue->session,queue->transit);
    }
  if (status<1) {
    (void) clm_alert(0,"%s Unable to delete inqueue session=<%s> transit='%d'",
			"gesdbc.c:dbc_delqueue,",queue->session,queue->transit);
    done=false;
    }
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure once time only when Clement is*/
/*	started. Purpose in to put back in queue*/
/*	transite request which where		*/
/*	"in progress" mode while clement	*/
/*	previously stopped.			*/
/*	inqueue list.				*/
/*						*/
/************************************************/
int dbc_updlostqueue(SQLFILE *sql)

{
#define	POSTLST	"UPDATE %s SET progress=%d WHERE progress=%d"

return sql_action(sql,POSTLST,INQUEUE,QWAIT,QPROG);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to set a new nexttime for	*/
/*	email in queue for a specific domain.	*/
/*	inqueue list.				*/
/*						*/
/************************************************/
int dbc_postponequeue(SQLFILE *sql,char *recdom,int secdelay)

{
#define	POSTQUE	"UPDATE %s SET next=NOW()+'%d sec' WHERE progress=%d AND recdom='%s'"

return sql_action(sql,POSTQUE,INQUEUE,secdelay,QWAIT,recdom);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if the proposed	*/
/*	domain is in fact a domain alias.	*/
/*						*/
/************************************************/
char *dbc_isitadomainalias(SQLFILE *sql,char *domain)

{
#define GETDOMAL	"SELECT * FROM %s WHERE domainaliased=%s"
char *domalias;
char *cleandom;
SQLRESULT *rs;

domalias=(char *)0;
cleandom=sql_cleanstr(sql,domain);
if ((rs=sql_tupple(sql,GETDOMAL,DOMAL,cleandom))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) 
    domalias=sql_strdup(sql_getvalue(sql,rs,0,"domainname"));
  (void) sql_dropresult(sql,rs);
  }
(void) free(cleandom);
return domalias;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to set list of quarantined 	*/
/*	email to extract queue when local user	*/
/*	is sending Email to a remote user.	*/
/*						*/
/************************************************/
void dbc_purgequar(SQLFILE *sql,char *originator,char *recipient)


{
#define	OPPU	"gesdbc.c:dbc_purgequar,"
#define	FPUR1	"session,recipient"
#define	PURGE	"INSERT INTO %s (%s) (SELECT %s FROM %s WHERE status=%d AND recipient=%s AND originator=%s)"

char *orgeml;
char *receml;

orgeml=sql_cleanstr(sql,originator);
receml=sql_cleanstr(sql,recipient);
if (sql_action(sql,PURGE,EXTRACT,FPUR1,FPUR1,EMLOGS,m_quarantine,orgeml,receml)<0) {
  char *cmd;
 
  (void) asprintf(&cmd,PURGE,EXTRACT,FPUR1,FPUR1,EMLOGS,m_quarantine,orgeml,receml);
  (void) clm_alert(0,"%s Unable to do SQL request=<%s> (bug?)",OPPU,cmd);
  (void) free(cmd);
  }
(void) free(receml);
(void) free(orgeml);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return true if the 	*/
/*	authenticated originator sent too many	*/
/*	outside its own group.			*/
/*						*/
/************************************************/
int dbc_toomanymail(SQLFILE *sql,char *authemail,int groupid,unsigned int *sent)

{
#define	OPOO	"gesdbc.c:dbc_toomanymail,"

#define	ISUSR	"%s %s WHERE username=%s"

int toomany;
int usrdta;
GRPTYP *grp;
PRFTYP *prf;
int phase;
int proceed;

toomany=false;
usrdta=0;
*sent=0;
prf=(PRFTYP *)0;
grp=(GRPTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting group id	*/
      if (groupid==0)
	groupid++;	/*default external user	*/
      break;
    case 1	:	/*getting group id<-usr	*/
      if ((usrdta=getgroupid(sql,authemail))>1) {
	groupid=usrdta;
	}
      break;
    case 2	:	/*# Email sentby auth	*/
      *sent=getcount(sql,authemail,groupid);
      break;
    case 3	:	/*getting user group	*/
      if ((grp=dbc_getgroupbyid(sql,groupid))==(GRPTYP *)0) {
        (void) clm_alert(0,"%s Unable to find group for user <%s>  and groupid '%d' (bug?)",OPOO,authemail,groupid);
	proceed=false;
	}
      break;
    case 4	:	/*getting user profile	*/
      if (grp!=(GRPTYP *)0) {	/*always	*/
	char *username;
	SQLRESULT *rs;

	username=sql_cleanstr(sql,authemail);
        if ((rs=sql_tupple(sql,ISUSR,SELFROM,USERS,username))!=(SQLRESULT *)0) {
          if (sql_nbrtupple(sql,rs)>0) {
            if ((usrdta=getcommonvalue(sql,authemail,"prfid",0))>0) {
	      /*user has its own profile!			*/
	      grp->prfid=usrdta;
	      }
	    }
          (void) sql_dropresult(sql,rs);
	  }
	(void) free(username);
	}
      if ((prf=dbc_getprofile(sql,grp->prfid))==(PRFTYP *)0) {
        (void) clm_alert(0,"%s Unable to find profile for user <%s>  and groupid/profid '%d/%d' (bug?)",
			    OPOO,authemail,groupid,grp->prfid);
	proceed=false;
        }
      grp=dbc_freegroup(grp);
      break;
    case 5	:	/*checking email maximum*/
      if (*sent>prf->maxperday)
	toomany=true;
      prf=dbc_freeprofile(prf);
      break;
    default	:	/*safe Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return toomany;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve the domain	*/
/*	activation level			*/
/*						*/
/************************************************/
DACTYP dbc_getactive(SQLFILE *sql,char *domain)

{
DACTYP active;
time_t upto;
char *cleandom;
SQLRESULT *rs;

active=da_unknown;
cleandom=sql_cleanstr(sql,domain);
if ((rs=sql_tupple(sql,SELDOM,SELFROM,DOMS,cleandom))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) {
    switch ((atoi)(sql_getvalue(sql,rs,0,"active"))) {
      case 0	:	/*domain is locked	*/
	active=da_lock;
	break;
      case 1	:	/*domain is accepted	*/
	active=da_good;
        upto=getuptograce(sql,rs,0);
        if ((upto>0)&&(upto<time((time_t *)0)))
	  active=da_lock; 	/*domain expired*/
	break;
      case 2	:	/*domain is to resend	*/
	active=da_resend;
	break;
      default	:	/*unexpected status	*/
	(void) clm_alert(0,"%s Unexpected active status '%d' on domain <%s>",
			   "gesdbc.c:dbc_getactive,",(atoi)(sql_getvalue(sql,rs,0,"active")),domain);
	break;
      }
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(cleandom);
return active;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to build a DSP structure from	*/
/*	the data collationned within emlogs	*/
/*	table.					*/
/*						*/
/************************************************/
DSPTYP *dbc_getsessiondsp(SQLFILE *sql,char *session,char *recipient)

{
DSPTYP *dsp;
LOGTYP *emlog;
int phase;
int proceed;

dsp=(DSPTYP *)0;
emlog=(LOGTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting log info	*/ 
      if ((emlog=dbc_getlogs(sql,session,recipient))==(LOGTYP *)0)
	phase=999;	/*no log? trouble	*/
      break;
    case 1	:	/*converting log to dsp	*/
      dsp=(DSPTYP *)calloc(1,sizeof(DSPTYP));
      dsp->spmlvl=NAN;
      dsp->spmlim=NAN;
      dsp->session=emlog->session;
      dsp->peerip=emlog->rmtip;
      dsp->efrom=emlog->efrom;
      dsp->hfrom=emlog->hfrom;
      dsp->grpin=emlog->grpin;
      dsp->subject=emlog->subject;
      break;
    case 2	:	/*cleaning log		*/
      emlog->session=(char *)0;
      emlog->rmtip=(char *)0;
      emlog->efrom=(char *)0;
      emlog->hfrom=(char *)0;
      emlog->subject=(char *)0;
      emlog=dbc_freelogs(emlog);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return dsp;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to store session log data	*/
/*	within the seslogs table.		*/
/*						*/
/************************************************/
int dbc_rptsession(SQLFILE *sql,char *session,char *seslog)

{
#define	OPRP	"gesdbc.c:dbc_rptsession,"
#define INSLOGS	"INSERT INTO %s (sesid,logs) values('%s',%s)"

int done;
char *cleanlogs;

done=true;
cleanlogs=sql_cleanstr(sql,seslog);
if (sql_action(sql,INSLOGS,SESLOGS,session,cleanlogs)<0) {
  (void) clm_alert(0,"%s Unable to do SQL request=<"INSLOGS"> (bug?)",
		     OPRP,SESLOGS,session,cleanlogs);
  done=false;
  }
(void) free(cleanlogs);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to setup a clean from data	*/
/*	for each recipient, this can be not	*/
/*	the same within the same email, if an	*/
/*	alias recipient expension is used.	*/
/*						*/
/************************************************/
void dbc_setinbehalf(RCPTYP **rcpts,char *inbehalf)

{
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) 
    rcpts[i]->inbehalf=strdup(inbehalf);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find out about the langid	*/
/*	related to a known user.		*/
/*						*/
/************************************************/
int dbc_getusrlang(SQLFILE *sql,char *username)

{
int lang;
PRFTYP *prf;

lang=DFLTLNG;
if ((prf=dbc_getemailprofile(sql,username))!=(PRFTYP *)0) {
  lang=prf->lang;
  prf=dbc_freeprofile(prf);
  }
return lang;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update user status, if	*/
/*	user is not existing, user record is	*/
/*	created.				*/
/*						*/
/************************************************/
int dbc_updateuser(SQLFILE *sql,USRTYP *usr)

{
#define	OPTE	"gesdbc.c:dbc_updateuse,"

#define UPDUSR	"UPDATE %s SET active=%d,upto=NOW() WHERE username=%s"
#define NEWUSR	"INSERT INTO %s (username,active,prfid,first) VALUES (%s,%d,%d,NOW())"

static LCKTYP lck1={USERS,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
register char *username;
int phase;
int proceed;

done=true;
username=sql_cleanstr(sql,usr->username);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is record existing?	*/
      if (sql_action(sql,UPDUSR,USERS,usr->active,username)==1) {
        phase=999;	/*record update done	*/
	}
      break;
    case 1	:	/*updating user profile	*/
      if (usr->prfid==0) {
        GRPTYP *grp;

	if ((grp=dbc_getgroup(sql,usr->username))!=(GRPTYP *)0) {
	  usr->prfid=grp->prfid;
	  grp=dbc_freegroup(grp);
	  }
	else {
          (void) clm_alert(0,"%s Unable to find user <%s> group (DB trouble?)",OPTE,usr->username);
	  }
	}
      break;
    case 2	:	/*setting a lock	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPTE,USERS);
        done=false;
        phase=999;	/*trouble trouble	*/
        }
      break;
    case 3      :       /*doing update again    */
      if (sql_action(sql,UPDUSR,USERS,usr->active,username)==1)
        phase++;	/*created in mid-time	*/
			/*update done		*/
      break;
    case 4	:	/*inserting new record  */
      if (done==true) {	/*always		*/
        if (sql_action(sql,NEWUSR,USERS,username,usr->active,usr->prfid)!=1) {
          (void) clm_alert(0,"%s Unable to create %s record for <%s>",OPTE,USERS,usr->username);
          (void) sql_unlock_rollback(sql);
          done=false;
          phase=999;	/*trouble trouble	*/
          }
	}
      break;
    case 5	:	/*unlocking table	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) clm_freestr(username);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract cron directive to	*/
/*	be done at the right time.		*/
/*						*/
/************************************************/
char **dbc_getcrontab(SQLFILE *sql,time_t curtime)

{
#define	TIMETZ	"TZ"	/*Time Zone		*/
#define	GETGRP	"SELECT * FROM %"

char **crontab;
GRPTYP **groups;

crontab=(char **)0;
if ((groups=dbc_getgrouplst(sql))!=(GRPTYP **)0) {
  char *mytz;
  int i;

  mytz=getenv(TIMETZ);
  for (i=0;groups[i]!=(GRPTYP *)0;i++) {
    char **crongrp;

    if (groups[i]->timezone!=(char *)0) {
      (void) setenv(TIMETZ,groups[i]->timezone,1);
      (void) tzset();
      }
    if ((crongrp=getcron(sql,groups[i]->groupid,curtime))!=(char **)0) {
      int j;

      for (j=0;crongrp[j]!=(char *)0;j++) {
	char **ptr;
        char *nptr;

	if ((ptr=getcronptr(crontab,crongrp[j]))==(char **)0) {
	  nptr=strdup(crongrp[j]);
          crontab=(char **)vec_addlstlst((void **)crontab,nptr);
	  ptr=crontab+vec_sizelstlst((void **)crontab);
	  ptr--;
	  }
        (void) asprintf(&nptr,"%s %d",*ptr,groups[i]->groupid);
	(void) free(*ptr);
	*ptr=nptr;
	}
      crongrp=(char **)vec_freelstlst((void **)crongrp,VECFREE);
      }
    }
  (void) unsetenv(TIMETZ); 
  if (mytz!=(char *)0)
    (void) setenv(TIMETZ,mytz,1);
  (void) tzset();
  groups=dbc_freegrouplst(groups);
  }
return crontab;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract info from a banned	*/
/*	(or trusted) bantrust table.		*/
/*	status about an IP			*/
/*						*/
/************************************************/
char **dbc_getbaninfo(SQLFILE *sql,char *ipnum)

{
char **info;
int pos;
SQLRESULT *rs;

info=(char **)0;
if ((rs=locateban(sql,ipnum,&pos))!=(SQLRESULT *)0) {
  char **lsts;

  if ((lsts=clm_splitstring(sql_getvalue(sql,rs,pos,"info"),'\n'))!=(char **)0) {
    char *str;
    int i;

    (void) asprintf(&str,"%s/%d",sql_getvalue(sql,rs,pos,"ipnum"),
			         atoi(sql_getvalue(sql,rs,pos,"mask")));
    info=(char **)vec_addlstlst((void **)info,(void *)str);
    for (i=0;lsts[i]!=(char *)0;i++) 
      info=(char **)vec_addlstlst((void **)info,lsts[i]);
    (void) asprintf(&str,".");
    info=(char **)vec_addlstlst((void **)info,(void *)str);
    (void) free(lsts);
    }
  (void) sql_dropresult(sql,rs);
  }
return info;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a specific flag	*/
/*	is set with the trusted table.		*/
/*						*/
/************************************************/
int dbc_gettrust(SQLFILE *sql,char *ipnum,int trsmask)

{
#define	SELTRT	"SELECT * FROM %s WHERE ipnum='%s' AND mask=%d"

int trustus;
AFNTYP *afnnum;
SQLRESULT *rs;

trustus=false;
afnnum=afn_getipnum(ipnum);
if ((rs=sql_tupple(sql,SELTRT,TRSTBL,ipnum,afnnum->afnmask))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) {
    if ((atoi(sql_getvalue(sql,rs,0,"status"))&trsmask)!=0)
      trustus=true;
    }
  (void) sql_dropresult(sql,rs);
  }
afnnum=afn_freeipnum(afnnum);
return trustus;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to flag a remote server as	*/
/*	trusting us.				*/
/*	status about an IP			*/
/*						*/
/************************************************/
int dbc_settrust(SQLFILE *sql,char *ipnum,int trsmask,int trsflag,time_t updt)

{
#define	OPTR	"gesdbc.c:dbc_settrust,"

#define UPDTRS	"UPDATE %s SET status=status%s(%s%d),updt='%s' WHERE ipnum='%s' AND mask='%d'"
#define NEWTRS	"INSERT INTO %s (status,iptype,ipnum,reverse,ipbin,mask,updt) VALUES (%d,%d,'%s','%s',%s,%d,'%s')"

static LCKTYP lck1={TRSTBL,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
register const char *strupdt;
const char *op1;
const char *op2;
AFNTYP *afnnum;
int phase;
int proceed;

done=true;
strupdt=sql_fromunixtime(sql,updt);
op1="|";
op2="";
if (trsflag==false) {
  op1="&";
  op2="~"; 
  }
afnnum=afn_getipnum(ipnum);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is record existing?	*/
      if (sql_action(sql,UPDTRS,TRSTBL,op1,op2,trsmask,strupdt,ipnum,afnnum->afnmask)==1) {
        phase=999;	/*record already present*/
	}
      break;
    case 1	:	/*setting a lock	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPTR,TRSTBL);
        done=false;
        phase=999;	/*trouble trouble	*/
        }
      break;
    case 2      :       /*doing update again    */
      if (sql_action(sql,UPDTRS,TRSTBL,op1,op2,trsmask,strupdt,ipnum,afnnum->afnmask)==1) 
        phase++;	/*created in mid-time	*/
      break;
    case 3	:	/*inserting new record  */
      if (done==true) {	/*always		*/
	char *ipbin;
	char *reverse;
	char **rvrlst;

	ipbin=sql_toipbin(sql,ipnum);
	reverse=NORVERS;
	if (((rvrlst=dns_getptr(afnnum))!=(char **)0)&&(rvrlst[0]!=(char *)0))
	  reverse=rvrlst[0];
        if (trsflag==false)
	  trsmask=0;
        if (sql_action(sql,NEWTRS,TRSTBL,trsmask,afnnum->afntype,ipnum,reverse,ipbin,afnnum->afnmask,strupdt)!=1) {
          (void) clm_alert(0,"%s Unable to create Trusted record for <%s/%d>",OPTR,ipnum,afnnum->afnmask);
          (void) sql_unlock_rollback(sql);
          done=false;
          phase=999;	/*trouble trouble	*/
          }
    	rvrlst=(char **)vec_freelstlst((void **)rvrlst,VECFREE);
        (void) free(ipbin);
	}
      break;
    case 4	:	/*unlocking table	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
afnnum=afn_freeipnum(afnnum);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract banned/trusted	*/
/*	status about an IP			*/
/*						*/
/************************************************/
int dbc_getban(SQLFILE *sql,char *ipnum)

{
#define	OPBA	"gesdbc.c:dbc_getban,"

#define	UPDBAN	"UPDATE %s SET last=NOW() WHERE ipnum='%s'"

int status;
SQLRESULT *rs;
int pos;

status=0;
if ((rs=locateban(sql,ipnum,&pos))!=(SQLRESULT *)0) {
  status=atoi(sql_getvalue(sql,rs,pos,"status"));
  (void) sql_action(sql,UPDBAN,BANTBL,sql_getvalue(sql,rs,pos,"ipnum"));
  (void) sql_dropresult(sql,rs);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to insert banned/trusted	*/
/*	status and info about an IP		*/
/*	Return the number of ip inserted.	*/
/*						*/
/************************************************/
int dbc_setban(SQLFILE *sql,char **iplists,int status,char *info)

{
#define	OPSB	"gesdbc.c:dbc_setban,"

#define INBAN	"INSERT INTO %s (status,iptype,ipnum,reverse,ipbin,mask,info,updt) VALUES (%d,%d,'%s','%s',%s,%d,'%s',now())"
#define UPDEML	"UPDATE %s SET status=%d WHERE rmtip='%s' AND (status=%d OR status=%d)"

static LCKTYP lck1={BANTBL,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int number;
int dbstatus;
int phase;
int proceed;

number=0;
dbstatus=getbanval(status,"dbc_setban");
if (strlen(info)>MAXINFO) {
  (void) clm_alert(0,"%s Info=<%s> size '%d' too big, now limited to '%d'",OPSB,info,strlen(info),MAXINFO);
  info[MAXINFO]='\000';
  }
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Locking data-base	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPSB,BANTBL);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*scanning list		*/
      if (iplists!=(char **)0) {
	int i;
        char **nlists;

	nlists=(char **)0;
	for (i=0;iplists[i]!=(char *)0;i++) {
	  if ((dbc_getban(sql,iplists[i]))==0) {
	    AFNTYP *afnnum;
	    char *reverse;
	    char **rvrlst;
	    char *ipbin;

	    afnnum=afn_getipnum(iplists[i]);
	    reverse=NORVERS;
	    if (((rvrlst=dns_getptr(afnnum))!=(char **)0)&&(rvrlst[0]!=(char *)0))
	      reverse=rvrlst[0];
	    ipbin=sql_toipbin(sql,iplists[i]);
	    if (sql_action(sql,INBAN,BANTBL,dbstatus,afnnum->afntype,iplists[i],reverse,ipbin,afnnum->afnmask,info)<0)
	      (void) clm_alert(0,"%s Unable to store %s/%d within %s data-base table (bug?)",
			         OPSB,iplists[i],afnnum->afnmask,BANTBL);
	    ipbin=clm_freestr(ipbin);
    	    rvrlst=(char **)vec_freelstlst((void **)rvrlst,VECFREE);
	    afnnum=afn_freeipnum(afnnum);
	    number++;
            nlists=(char **)vec_addlstlst((void **)nlists,(void *)iplists[i]);
	    continue;
	    }
          iplists[i]=clm_freestr(iplists[i]);
	  }
	if (nlists!=(char **)0) {
	  (void) memcpy(iplists,nlists,sizeof(char *)*(vec_sizelstlst((void **)nlists)+1));
	  (void) free(nlists);
	  }
	}
      break;
    case 2	:	/*Unlock data-base	*/
      (void) sql_unlock_commit(sql);
      break;
    case 3	:	/*Updating emlogs	*/
      if ((iplists!=(char **)0)&&(status==RBANNED)) {
	int i;
	
	for (i=0;iplists[i]!=(char *)0;i++) {
	  if (sql_action(sql,UPDEML,EMLOGS,(int)m_ipbanned,iplists[i],(int)m_noaddress,(int)m_rejected)<0)
	    (void) clm_alert(0,"%s Unable to update %s within %s data-base table (bug?)",
			        OPSB,iplists[i],EMLOGS);
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return number;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove an entry from	*/
/*	the banned list.			*/
/*						*/
/************************************************/
int dbc_delban(SQLFILE *sql,char *ipnum,int mask)

{
#define DELBAN	"DELETE FROM %s WHERE ipnum='%s' AND mask=%d"

return sql_action(sql,DELBAN,BANTBL,ipnum,mask);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update banned/trusted	*/
/*	status and info about a specific IP.	*/
/*						*/
/************************************************/
int dbc_updban(SQLFILE *sql,char *ipnum,int mask,int status,char *info)


{
#define	OPUP	"gesdbd.c:dbc_updban,"

#define UPIBAN	"UPDATE %s SET updt=NOW(),status=%d,info='%s'  WHERE ipnum='%s' AND mask=%d"
#define NEWBAN	"INSERT INTO %s (updt,status,iptype,ipnum,reverse,ipbin,mask,info) VALUES (NOW(),%d,%d,'%s','%s',%s,%d,'%s')"

static LCKTYP lck1={BANTBL,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
AFNTYP *afnnum;
int phase;
int proceed;

done=true;	/*Let say eveything fine	*/
afnnum=afn_getipnum(ipnum);
(void) afn_maskip(afnnum,mask);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is record existing?	*/
      if (sql_action(sql,UPIBAN,BANTBL,status,info,ipnum,mask)==1)
        phase=999;	/*record already present*/
      break;
    case 1	:	/*setting a lock	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPUP,BANTBL);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*doing update again	*/
      if (sql_action(sql,UPIBAN,BANTBL,status,info,ipnum,mask)==1)
	phase++;
      break;
    case 3	:	/*inserting new record	*/
      if (done==true) {	/*always		*/
	char *reverse;
	char **rvrlst;
	char *ipbin;

	reverse=NORVERS;
	if (((rvrlst=dns_getptr(afnnum))!=(char **)0)&&(rvrlst[0]!=(char *)0))
	  reverse=rvrlst[0];
	ipbin=sql_toipbin(sql,ipnum);
        if (sql_action(sql,NEWBAN,BANTBL,status,afnnum->afntype,ipnum,reverse,ipbin,mask,info)!=1) {
          (void) clm_alert(0,"%s Unable to create Banned record for <%s/%d>",OPUP,ipnum,mask);
          (void) sql_unlock_rollback(sql);
	  done=false;
	  phase=999;	/*trouble trouble	*/
	  }
    	rvrlst=(char **)vec_freelstlst((void **)rvrlst,VECFREE);
	ipbin=clm_freestr(ipbin);
        }
      break;
    case 4	:	/*unlocking table	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    } 
  phase++;
  }
afnnum=afn_freeipnum(afnnum);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to build the ratebantrust	*/
/*	data base table.			*/
/*						*/
/************************************************/
int dbc_buildbantrust(SQLFILE *sql,int minimal)

{
#define	OPBU	"gesdbd.c:dbc_buildbantrust,"

#define	DELRAT	"DELETE FROM %s"
#define	POLRAT	"INSERT INTO %s (good,total,ipnum) SELECT good,total,ipnum FROM (SELECT "
#define	POLRAT1	"count(nullif(status<%d,false)) AS total,"
#define	POLRAT2 "count(nullif(status=%d OR status=%d OR status=%d OR status=%d,false)) AS good,"
#define	POLRAT3	"rmtip as ipnum FROM emlogs GROUP BY ipnum) AS collect WHERE total>=%d"
#define	TRIMRAT	"DELETE FROM %s WHERE total=0.0"
#define	CALRAT	"UPDATE %s SET ratio=(total-good)/total"

register int done;
register int phase;
register int proceed;

done=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*begin transaction	*/
      if (sql_lock_action(sql)<0) {
        (void) clm_alert(0,"%s Unable to dbd_locksoft",OPBU);
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*preparing table	*/
      if (sql_action(sql,DELRAT,RATEBAN)<0) {
        (void) clm_alert(0,"%s Unable to delete table <%s> contents (bug?)",OPBU,RATEBAN);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*building table	*/
      if (sql_action(sql,POLRAT POLRAT1 POLRAT2 POLRAT3,RATEBAN,
			 (int)m_ipbanned,(int)m_acceptable,(int)m_toforward,(int)m_delivered,(int)m_forwarded,
			 minimal)<0) {
        (void) clm_alert(0,"%s Unable to build table <%s> contents (bug?)",OPBU,RATEBAN);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*adjusting table	*/
      if (sql_action(sql,TRIMRAT,RATEBAN)<0) {
        (void) clm_alert(0,"%s Unable to remove null total from table <%s> contents (bug?)",OPBU,RATEBAN);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*computing ratio table	*/
      if (sql_action(sql,CALRAT,RATEBAN)<0) {
        (void) clm_alert(0,"%s Unable to compute ratio on table <%s> contents (bug?)",OPBU,RATEBAN);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 5	:	/*everything fine	*/
      done=true;
      (void) sql_unlock_commit(sql);
      proceed=false;
      break;
    default	:	/*SAFE Guard (trouble)	*/
      (void) sql_unlock_rollback(sql);
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract remote IP on which	*/
/*	the trust level is above or below a	*/
/*	rate treshold.				*/
/*						*/
/************************************************/
char **dbc_extractrate(SQLFILE *sql,float rate,int trusted)

{
#define	SELRAT	"SELECT * FROM %s WHERE ratio%c%3.2f ORDER BY ipnum %s"

char **rates;
char sel;
char *way;
SQLRESULT *rs;

rates=(char **)0;
sel='>';
way="desc";
if (trusted==true) {
  sel='<';
  way="asc";
  }
if ((rs=sql_tupple(sql,SELRAT,RATEBAN,sel,rate,way))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;

    for (i=0;i<nbr;i++) {
      const char *ptr;

      if ((ptr=sql_getvalue(sql,rs,i,"ipnum"))==(char *)0)
	continue;
      rates=(char **)vec_addlstlst((void **)rates,(void *)strdup(ptr));
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return rates;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to scan banned data-base and	*/
/*	agregate IP to make ip group.		*/
/*						*/
/************************************************/
int dbc_agregate(SQLFILE *sql,int netype)

{
#define	OPAG	"gesdbc.c:dbc_agregate,"

#define	MNGBAN	"UPDATE %s SET mask=%d WHERE ipnum='%s'"

#define	MXLIST	500	/*list batch size	*/

int isok;
int mlim;
int mask;
int i;

isok=false;
mlim=16;
mask=32;
if (netype==AF_INET6) {
  mlim=48;
  mask=128;
  }
for (i=0;i<mlim;i++,mask--) {
  int offset;
  char **iplists;

  offset=0;
  while ((iplists=listbanned(sql,netype,mask,offset,MXLIST))!=(char **)0) {
    int j;
    AFNTYP *af1;
    char foundip[80];

    (void) strcpy(foundip,iplists[0]);
    af1=afn_getipnum(foundip);
    offset++;
    for (j=1;iplists[j]!=(char *)0;j++,offset++) {
      AFNTYP *af2;

      (void) strcpy(foundip,iplists[j]);
      af2=afn_getipnum(foundip);
      if (afn_cmpipnum(af1,af2,mask-1)==true) {
	if (sql_action(sql,MNGBAN,BANTBL,mask-1,iplists[j-1])<0) 
          (void) clm_alert(0,"%s Unable to update <%s> to mask='%d'",OPAG,iplists[j-1],mask-1);
	if (dbc_delban(sql,iplists[j],mask)<0)
          (void) clm_alert(0,"%s Unable to delete <%s> mask='%d'",OPAG,iplists[j],mask);
	}
      af1=afn_freeipnum(af1);
      af1=af2;
      }
    af1=afn_freeipnum(af1);
    iplists=(char **)vec_freelstlst((void **)iplists,VECFREE);
    }
  }
return isok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the list of remote	*/
/*	server trusting us.			*/
/*						*/
/************************************************/
TRUDAT **dbc_gettrustingus(SQLFILE *sql)


{
#define	DELTRU	"DELETE FROM %s WHERE ipnum IN (SELECT binding FROM %s)"
#define	SELTRU	"SELECT * FROM %s WHERE status&%d!=0"

TRUDAT **trulsts;
SQLRESULT *rs;

trulsts=(TRUDAT **)0;
/*Removing not needed trusting; No need to send banned info to myself	*/
(void) sql_action(sql,DELTRU,TRSTBL,BINDING);
if ((rs=sql_tupple(sql,SELTRU,TRSTBL,TRSTUS))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;

    for (i=0;i<nbr;i++) {
      TRUDAT *tru;
	
      tru=calloc(1,sizeof(TRUDAT));
      tru->ipnum=sql_strdup(sql_getvalue(sql,rs,i,"ipnum"));
      tru->ipnum=afn_unformatip(tru->ipnum);
      tru->updt=sql_tounixtime(sql,sql_getvalue(sql,rs,i,"updt"));
      trulsts=(TRUDAT **)vec_addlstlst((void **)trulsts,(void *)tru);
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return trulsts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a 	*/
/*	trustingus list.			*/
/*						*/
/************************************************/
TRUDAT **dbc_freetrustingus(TRUDAT **trulsts)

{
if (trulsts!=(TRUDAT **)0) {
  int i;

  for (i=0;trulsts[i]!=(TRUDAT *)0;i++) {
    trulsts[i]->ipnum=clm_freestr(trulsts[i]->ipnum);
    (void) free(trulsts[i]);
    }
  (void) free(trulsts);
  trulsts=(TRUDAT **)0;
  }
return trulsts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect a list of banned	*/
/*	which have been updated after upto	*/
/*						*/
/************************************************/
BANTYP **dbc_getbanlist(SQLFILE *sql,time_t *upto)

{
#define BANLIM	200
#define	SELBANN	"SELECT * FROM %s WHERE updt>'%s' ORDER BY updt ASC LIMIT %d"

BANTYP **banlsts;
register const char *YDAY;
SQLRESULT *rs;

banlsts=(BANTYP **)0;
YDAY=sql_fromunixtime(sql,*upto);
if ((rs=sql_tupple(sql,SELBANN,BANTBL,YDAY,BANLIM))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;

    *upto=sql_tounixtime(sql,sql_getvalue(sql,rs,nbr-1,"updt"));
    for (i=0;i<nbr;i++) {
      BANTYP *banloc;
      register char *info;
      register char **split;

      banloc=(BANTYP *)calloc(1,sizeof(BANTYP)); 
      banloc->mask=atoi(sql_getvalue(sql,rs,i,"mask"));
      banloc->banip=sql_strdup(sql_getvalue(sql,rs,i,"ipnum"));
      banloc->infos=calloc(3,sizeof(char *));
      banloc->infos[0]=afn_unformatip(sql_strdup(sql_getvalue(sql,rs,i,"ipnum")));
      (void) asprintf(&(banloc->infos[1]),"/%s",sql_getvalue(sql,rs,i,"mask"));
      info=sql_getvalue(sql,rs,i,"info");
      if ((info!=(char *)0) && ((split=clm_splitstring(info,'\n'))!=(char **)0)) {
	int j;

	for (j=0;split[j]!=(char *)0;j++) {
          banloc->infos=(char **)vec_addlstlst((void **)banloc->infos,(void *)split[j]);
	  }
        banloc->infos=(char **)vec_addlstlst((void **)banloc->infos,(void *)(strdup(".")));
 	(void) free(split);
	}
      banlsts=(BANTYP **)vec_addlstlst((void **)banlsts,(void *)banloc);
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return banlsts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	BANTYP list.				*/
/*						*/
/************************************************/
BANTYP **dbc_freebanlsts(BANTYP **banlsts)

{
if (banlsts!=(BANTYP **)0) {
  int i;

  for (i=0;banlsts[i]!=(BANTYP *)0;i++) {
    banlsts[i]->banip=clm_freestr(banlsts[i]->banip);
    banlsts[i]->infos=(char **)vec_freelstlst((void **)banlsts[i]->infos,VECFREE);
    (void) free(banlsts[i]);
    }
  (void) free(banlsts);
  banlsts=(BANTYP **)0;
  }
return banlsts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find the IDPATH extracted	*/
/*	a foreign email (an email not related to*/
/*	ISP).					*/
/*						*/
/************************************************/
char *dbc_getidpath(SQLFILE *sql,char *multipath)

{
#define	SELIDP	"SELECT * FROM %s WHERE e_email=%s"

char *idpath;
char *e_email;
SQLRESULT *rs;

idpath=(char *)0;
e_email=sql_cleanstr(sql,multipath);
if ((rs=sql_tupple(sql,SELIDP,MPATH,e_email))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) 
    idpath=sql_strdup(sql_getvalue(sql,rs,0,"i_email"));
  (void) sql_dropresult(sql,rs);
  }
(void) free(e_email);
return idpath;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a recent Email	*/
/*	was marked with specifig Msgid.		*/
/*						*/
/************************************************/
int dbc_findmsgid(SQLFILE *sql,char *session,char *msgid)

{
#define	LOOKID	"SELECT * FROM %s WHERE msgid=%s AND (status=%d OR status=%d OR status=%d) AND stamp>NOW()+'-%d sec' ORDER BY stamp DESC"
#define LOOKRNG	14400	/*look back 4 hours	*/

int status;
char *cleanid;
SQLRESULT *rs;

status=false;
cleanid=sql_cleanstr(sql,msgid);
if ((rs=sql_tupple(sql,LOOKID,EMLOGS,cleanid,m_forwarded,m_toforward,m_mustfwd,LOOKRNG))!=(SQLRESULT *)0) {
  int nbr;

  nbr=sql_nbrtupple(sql,rs);
  if (nbr>0) {
    int i;

    for (i=0;i<nbr;i++) {
      if (strcmp(sql_getvalue(sql,rs,i,"session"),session)!=0) {
        status=true;
	break;
	}
      }
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(cleanid);
return status;
}
