/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	MODTST:					*/
/*	Module doing components validation test	*/
/*						*/
/************************************************/
#include	<unistd.h>
#include	<signal.h>
#include	<stdio.h>
#include	<errno.h>
#include	<time.h>
#include	<string.h>
#
#include	"lowtyp.h"
#include	"utlprc.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"unieml.h"
#include	"unisql.h"
#include	"unijrl.h"
#include	"lvlcnf.h"
#include	"lvlpro.h"
#include	"gesipc.h"
#include	"gestls.h"
#include	"gesclm.h"
#include	"lvleml.h"

static char *status[]={"not successful","fully successful"};
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to write some test sequence to	*/
/*	journal.				*/
/*						*/
/************************************************/
static void writejournal(int numjrl,int mxline)

{
register int i;


for (i=0;i<mxline;i++) {
  (void) usleep(random()%100000);
  (void) jrl_journalling(false,numjrl,"Test journalling on jrl=%d "
				      "line %d (pid=%05d)",
				      numjrl,i+1,getpid());
  }
(void) jrl_journalling(true,numjrl,"");
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to check if journal function	*/
/*	are properly working			*/
/*						*/
/************************************************/
static int tstjournal()

{
#define	APPTEST	"testing"
#define MXTST MXJRL
#define	MXCHLD 5
#define	MXLINE 10

int itsok;
int success;
pid_t child[MXCHLD];
register int i,j;

itsok=0;
success=false;
(void) fprintf(stdout,"Begining Journal test\nPlease wait few seconds\n");
(void) srandom(MXTST);
for (j=0;j<MXTST;j++) {
  const char *jrlname;
  char apptest[50];

  (void) snprintf(apptest,sizeof(apptest),"%s%d",APPTEST,j);
  (void) clm_setjournal(APJRL+j,d_log,apptest);
  if ((jrlname=jrl_getjourname(APJRL+j))!=(char *)0) {
    (void) unlink(jrlname);
    }
  for (i=0;i<MXCHLD;i++) {
    switch (child[i]=fork()) {
      case -1	:
        (void) fprintf(stderr,"tst_journal, Unable to fork (error=<%s>)",
			       strerror(errno));
        break;
      case 0	:		/*sub process	*/
        (void) writejournal(APJRL+j,MXLINE);
        (void) exit(0);
        break;
      default	:		/*main process	*/
        (void) usleep(random()%1000);
        break;
      }
    }
  while (success==false) {
    register int chl;

    success=true;
    (void) sleep(1);
    (void) prc_nozombie();
    for (chl=0;chl<MXCHLD;chl++) {
      if (prc_checkprocess(child[chl])==true) {
        success=false;
	}
      }
    }
  success=false;
  if ((jrlname=jrl_getjourname(APJRL+j))!=(char *)0) {
    FILE *fichier;
    int numline;

    numline=0;
    if ((fichier=fopen(jrlname,"r"))!=(FILE *)0) {
      char line[300];

      while (fgets(line,sizeof(line),fichier)!=(char *)0) 
        numline++;
      (void) fclose(fichier);
      }
    if (numline==(MXLINE*MXCHLD)) {
      itsok++;
      }
    }
  }
(void) jrl_closejournals();
success=false;
if (itsok==MXTST)
  success=true;
(void) fprintf(stdout,"Journal test is <%s>\n",status[success]);
return success;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to send a request and wait	*/
/*	for a reply.				*/
/*						*/
/************************************************/
static void caller(BPTR *bptr,int iter,pid_t cherpid)

{
#define OPCA	"modtst.c:caller,"
#define	MXRQ	4000

u_long pool;

(void) srand(iter+11);
if ((pool=ipc_getpool(bptr,cherpid))==(u_long)0)
   (void) clm_core_dump("Unable to get cruncher pool value");
(void) prc_settitle("%s Caller iteration=%02d pool=%lu",appname,iter,pool);
if (ipc_subscribe(bptr,(u_long)0,true)==true) {
  time_t last;
  u_long many;
  u_long cpt;
  u_long rply;
  TIMESPEC timeout;
  char *msgref;
  char *msg;

  last=time((time_t *)0);
  many=(u_long)0;
  cpt=(u_long)0;
  rply=(u_long)0;
  while (sigterm==false) {
    #define MINTAILLE    3
    #define MXTAILLE	20 
    int taille;
    char *corps;
    int s;

    taille=(rand()%MXTAILLE)+MINTAILLE;
    corps=(char *)calloc(taille,sizeof(char));
    (void) memset(corps,'-',taille-1);
    corps[0]='A';
    corps[taille-1]='Z';
    for (s=1;s<taille-1;s++) {
      if ((s%50)==0)
	corps[s]='\n';
      }
    (void) asprintf(&msg,"%lu %05d taille=%u %s",cpt,getpid(),taille,corps);
    timeout.tv_sec=5;
    timeout.tv_nsec=0;
    if ((msgref=ipc_postmsg(bptr,&timeout,pool,msg))!=(char *)0) {
      char *got;
      time_t cur;
   
      timeout.tv_sec=5;
      timeout.tv_nsec=0;
      cpt++;
      if ((got=ipc_getreply(bptr,&timeout,msgref))!=(char *)0) {
	int j;

	for (j=0;j<3;j++) {
	  switch (j) {
	    case  0	:
	      if (strstr(got,"OK")!=got) {
                (void) clm_alert(0,"%s NOT OK reply <%s> to msg=<%s>",
			            OPCA,getpid(),got,msg);
	        j=3;
		}
	      break;
	    case  1	:
	      if (strcmp(got+3,msg)!=0) {
                (void) clm_alert(0,"%s rply (=%lu/t=%d) <%s>!=msg=<%s>",
			            OPCA,cpt,strlen(got+3),got,msg);
	        j=3;
		}
	      break;
	    case  2	:
	      many+=strlen(msg);
	      rply++;
	      break;
            }
	  }
        (void) free(got);
	cur=time((time_t)0);
	if (cur>last) {
	  (void) prc_settitle("%s Caller iteration=%02d pool=%lu cpt=%08lu",
			      appname,iter,pool,cpt);
	  last=cur;
	  }
	}
      else {
        (void) clm_alert(0,"(%05d) NO reply tsec='%lu':tnsec='%lu' to <%s>",
			    getpid(),
			    timeout.tv_sec,timeout.tv_nsec,msg);
	}
      (void) free(msgref);
      (void) free(corps);
      (void) free(msg);
      }
    else
      (void) clm_alert(0,"%s unable post msg=<%s>",OPCA,msg);
    }
  timeout.tv_sec=1;
  timeout.tv_nsec=0;
  if ((pool=ipc_getpool(bptr,getppid()))!=(u_long)0) {
    (void) asprintf(&msg,"%d %lu %lu %lu",getpid(),cpt,rply,many);
    if ((msgref=ipc_postmsg(bptr,&timeout,pool,msg))!=(char *)0) {
      (void) free(msgref);
      }
    (void) free(msg);
    }
  else
    (void) clm_alert(0,"Caller process %05d can't report to master",getpid());
  (void) ipc_subscribe(bptr,(u_long)0,false);
  }
else 
  (void) clm_alert(0,"Caller process %05d unable to subscribe",getpid());
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to wait for IPC message, then	*/
/*	return a value to the proper caller	*/
/*						*/
/************************************************/
static void cruncher(BPTR *bptr,u_int iter,u_long pool)

{
#define	OPTR	"modtst.c:cruncher,"

if (ipc_subscribe(bptr,pool,true)==true) {
  char *msg;
  char *msgref;
  TIMESPEC timeout;
  u_long cpt;
  time_t last;

  msg=(char *)0;
  msgref=(char *)0;
  cpt=0;
  last=time((time_t *)0);
  while (sigterm==false) {
    char *got;
    time_t cur;
    u_long millisec;

    timeout.tv_sec=1;
    timeout.tv_nsec=0;
    if ((got=ipc_getpost(bptr,&timeout,&msgref,&millisec))!=(char *)0) {
      char *reply;

      (void) asprintf(&reply,"OK %s",got);
      timeout.tv_sec=3;
      timeout.tv_nsec=0;
      if (ipc_postreply(bptr,&timeout,msgref,reply)==(u_long)0)
	(void) clm_alert(0,"%s Unable to post reply=<%s>",OPTR,reply);
      (void) free(reply);
      (void) free(msgref);
      (void) free(got);
      cpt++;
      }
    cur=time((time_t)0);
    if (cur>last) {
      (void) prc_settitle("%s Crunch iteration=%02d pool=%d cpt=%08lu",
			   appname,iter,pool,cpt);
      last=cur;
      }
    }
  (void) asprintf(&msg,"%d %d %lu",getpid(),iter,cpt);
  timeout.tv_sec=1;
  timeout.tv_nsec=0;
  if ((pool=ipc_getpool(bptr,getppid()))!=(u_long)0) {
    if ((msgref=ipc_postmsg(bptr,&timeout,pool,msg))!=(char *)0) {
      (void) free(msgref);
      }
    (void) free(msg);
    }
  else
    (void) clm_alert(0,"Cruncher process %05d can't report to master",getpid());
  (void) ipc_subscribe(bptr,pool,false);
  }
else
  (void) clm_alert(0,"%s, process %05d Unable to subscribe",OPTR,getpid());
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to start 'traitement' subprocess*/
/*						*/
/************************************************/
static void gescrunchers(BPTR *bptr,u_long pool,int num,
			 pid_t *trait,int start)

{
int i;

for (i=0;i<num;i++) {
  if (start==true) {
    switch (trait[i]=fork()) {
      case -1	:
        (void) clm_alert(0,"Unable to start traitement process");
	break;
      case  0	:	/*child process		*/
	(void) cruncher(bptr,i,pool);
	(void) exit(0);
	break;
      default	:	/*main process		*/
	break;
      }
    }
  else
    (void) kill(trait[i],SIGTERM);
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to start 'caller' subprocess	*/
/*						*/
/************************************************/
static void gescallers(BPTR *bptr,pid_t cherpid,int num,pid_t *trait,int start)

{
int i;

for (i=0;i<num;i++) {
  if (start==true) {
    switch (trait[i]=fork()) {
      case -1	:
        (void) clm_alert(0,"Unable to start caller process");
	break;
      case  0	:	/*child process		*/
	(void) caller(bptr,i,cherpid);
	(void) exit(0);
	break;
      default	:	/*main process		*/
	break;
      }
    }
  else
    (void) kill(trait[i],SIGTERM);
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to check if message exchange are*/
/*	properly working in single process.	*/
/*						*/
/************************************************/
static int singleipc()

{
#define	MXPILE	3
#define	MXMESS 10

int success;
BPTR *bptr;
u_long pool;
int i;
int phase;
int proceed;
char *msg[MXPILE];
char *msgref[MXPILE];

success=false;
bptr=(BPTR *)0;
pool=(u_long)0;
(void) memset(msg,'\000',sizeof(char *)*MXPILE);
(void) memset(msgref,'\000',sizeof(char *)*MXPILE);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*getting the memory	*/
      if ((bptr=ipc_getbox(MXMESS,2))==(BPTR *)0) {
        (void) clm_alert(0,"Unable to assigne shared memory");
	proceed=false;
	success=false;
	}
      break;
    case   1	:	/*getting a pool	*/
      pool=ipc_openpool(bptr);
      if (ipc_subscribe(bptr,pool,true)==false)
	phase=999;
      break;
    case   2	:	/*sending MXPILE msg	*/
      for (i=0;i<MXPILE;i++) {
	TIMESPEC timeout;

	(void) asprintf(&(msg[i]),"%05d %d  messages",getpid(),i);
        timeout.tv_sec=1;
        timeout.tv_nsec=0;
	msgref[i]=ipc_postmsg(bptr,&timeout,pool,msg[i]);
        } 
      break;
    case   3	:	/*getting back msg	*/
      success=true;
      for (i=0;i<MXPILE;i++) {
	char *got;
	TIMESPEC timeout;

        timeout.tv_sec=1;
        timeout.tv_nsec=0;
        if ((got=ipc_getreply(bptr,&timeout,msgref[i]))!=(char *)0) {
	  if (strcmp(got,msg[i])!=0)
	    success=false;
          (void) free(got);
	  }
	}
      break;
    case   4	:	/*cleaning memory	*/
      for (i=0;i<MXPILE;i++) {
        (void) free(msg[i]);
        (void) free(msgref[i]);
	}
      break;
    case   5	:	/*unsubscribing		*/
      (void) ipc_subscribe(bptr,(u_long)0,false);
      pool=ipc_closepool(bptr,pool);
      break;
    default	:	/*SAFE guard		*/
      bptr=ipc_dropbox(bptr);
      proceed=false;
      break;
    }
  phase++;
  }
if (success!=true)
  (void) clm_alert(0,"Single thread message NOT working");
return success;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to check if message exchange are*/
/*	properly working multiprocess mode.	*/
/*						*/
/************************************************/
static int multiipc()

{
#define	IPCDURE	    10 
#define	MXCRUNCHER   5 
#define MXCALLER    20
#define	MXBOX      500

int success;
BPTR *bptr;
u_long start;
u_long stop;
u_long totsent;
u_long totrply;
u_long totmany;
u_long totcrunched;
u_long pool;
TIMESPEC timer;
pid_t crunchers[MXCRUNCHER];
pid_t callers[MXCALLER];
time_t last;
int i;
int proceed;
int phase;


success=true;
bptr=(BPTR *)0;
start=(u_long)0;
stop=(u_long)0;
totsent=(u_long)0;
totrply=(u_long)0;
totmany=(u_long)0;
totcrunched=(u_long)0;
pool=(u_long)0;
last=time((time_t *)0);
(void) prc_settitle("%s modtest client=%d cruncher=%d ellapse time=%lu ms",
			appname,MXCALLER,MXCRUNCHER,0);
sigint=false;
proceed=true;
phase=0;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*setting the share memory	*/
      if ((bptr=ipc_getbox(MXBOX,MXCRUNCHER+MXCALLER+2))==(BPTR *)0) {
        (void) clm_alert(0,"Unable to assigne shared memory");
	proceed=false;
	success=false;
	}
      break;
    case  1	:	/*assigning a SIGTERM handler	*/
      (void) clm_alert(0,"Starting Inter process Exchange test, (%d crunchers, %d callers)",
			  MXCRUNCHER,MXCALLER);
      (void) clm_settrap(true);
      if ((pool=ipc_openpool(bptr))==(u_long)0)
	phase=999;
      break;
    case  2	:	/*starting du sub process	*/
      (void) gescrunchers(bptr,pool,MXCRUNCHER,crunchers,true);
      (void) gescallers(bptr,crunchers[0],MXCALLER,callers,true);
      break;
    case  3	:	/*waiting for convenient time	*/
      if (ipc_subscribe(bptr,(u_long)0,true)==false)
 	(void) clm_core_dump("Unable to subscribe main process to IPC");
      start=clm_getmillisec();
      if (debug==0)
        (void) clm_alert(0,"Wait '%d' seconds for test completion",IPCDURE);
      else
        (void) clm_alert(0,"Enter CTRL-C to stop test");
      while (sigint==false) {
	time_t curtime;

	curtime=time((time_t)0);
	timer.tv_sec=1;
	timer.tv_nsec=0;
	while ((nanosleep(&timer,&timer)!=0)&&(sigint==false));
	if (last<curtime) {
          (void) prc_settitle("%s modtest client=%d cruncher=%d time=%8.3f Sec",
			       appname,MXCALLER,MXCRUNCHER,
			       ((float)(clm_getmillisec()-start))/1000);
          (void) ipc_cleanbox(bptr,20);
	  last=curtime;
	  }
	if (debug==0) {
	  if ((clm_getmillisec()-start)>(IPCDURE*1000)) {
	    sigint=true;
	    }
	  }
        }
      (void) clm_alert(0,"Ending test and starting to collect data");
      break;
    case  4	:	/*sending TERM signal to caller	*/
      stop=clm_getmillisec();
      for (i=0;i<MXCALLER;i++) {
        (void) kill(callers[i],SIGTERM);
	}
      break;
    case  5	:	/*waiting for caller report	*/
      for (i=0;i<MXCALLER;i++) {
        char *got;
	char *msgref;
	u_long millisec;

        timer.tv_sec=2;
        timer.tv_nsec=0;
        if ((got=ipc_getpost(bptr,&timer,&msgref,&millisec))!=(char *)0) {
	  int call;
	  u_long sent;
	  u_long replied;
	  u_long many;

	  call=0;
	  sent=(u_long)0;
	  replied=(u_long)0;
	  many=(u_long)0;
	  if (sscanf(got,"%d %lu %lu %lu",&call,&sent,&replied,&many)==4) {
	    totsent+=sent;
	    totrply+=replied;
	    totmany+=many;
	    }
	  (void) free(msgref);
	  (void) free(got);
	  }
	}
      break;
    case  6	:	/*sending TERM signal to cruncher*/
      for (i=0;i<MXCRUNCHER;i++)
        (void) kill(crunchers[i],SIGTERM);
      break;
    case  7	:	/*waiting for cruncher report	*/
      for (i=0;i<MXCRUNCHER;i++) {
        char *got;
	char *msgref;
	u_long millisec;

        timer.tv_sec=2;
        timer.tv_nsec=0;
        if ((got=ipc_getpost(bptr,&timer,&msgref,&millisec))!=(char *)0) {
	  int call;
	  int iter;
	  u_long replied;

	  call=0;
	  iter=0;
	  replied=(u_long)0;
	  if (sscanf(got,"%d %d %lu",&call,&iter,&replied)==3) {
            (void) clm_alert(0,"Cruncher iter=%d sent %lu replies",
			        iter,replied);
	    totcrunched+=replied;
	    }
	  (void) free(msgref);
	  (void) free(got);
	  }
	}
      (void) ipc_subscribe(bptr,(u_long)0,false);
      pool=ipc_closepool(bptr,pool);
      break;
    default	:	/*safe guard			*/
      (void) clm_alert(0,"Reporting data statistic");
      bptr=ipc_dropbox(bptr);/*releasing shared memory	*/
      (void) clm_alert(0,"Sent %lu messages got %lu messages %lu crunched",
			  totsent,totrply,totcrunched);
      if (totsent==totrply) {
	float duration;

	duration=((float)stop-start)/1000.0;
        (void) clm_alert(0,"IPC speed %6.2f messages/sec (%lu Bytes) "
			   "for %8.3f sec",
			   (float)totsent/duration,totmany,duration);
	}
      (void) clm_settrap(false);
      (void) clm_alert(0,"Inter process Exchange test completed");
      proceed=false;
      break;
    }
  phase++;
  }
return success;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to check if message exchange are*/
/*	properly first in single process mode,	*/
/*	then in multiprocess mode..		*/
/*						*/
/************************************************/
static int tstipc()

{
int succes;

(void) clm_setjournal(APJRL,d_log,appname);
if ((succes=singleipc())==true)
  succes=multiipc();
(void) jrl_closejournals();
return succes;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to handle TSL exchanges		*/
/*						*/
/************************************************/
static void tlsaction(MXCTYP *mxc,LSTTYP *channel,int num)

{
#define	OPTL	"modtst.c:tlsaction,"

int cycle;

cycle=1;
while (sigterm==false) {
  CTTYP *contact;

  (void) prc_settitle("%s waiting on [%s:%d] (%02d/%02d/%04d)",
                      appname,channel->binding->ip,
                      channel->binding->port,num+1,
		      channel->binding->iteration,cycle);
  if ((contact=eml_acceptcontact(channel))!=(CTTYP *)0) {
    int proceed;

    proceed=true;
    while (proceed==true) {
      SCODTYP scode;

      switch (scode=eml_docontact(contact)) {
	case s_unknown	:
	  (void) clm_alert(0,"%s Unknown received command '%d'",OPTL,scode);
				/*NO BREAK	*/
	case s_timeout	:	/*contact lost		*/
	case s_quit	:
	  proceed=false;
	  break;
	default		:
	  (void) clm_alert(0,"%s Unexpected received command '%d' mxc='%#x'",OPTL,scode,mxc);
	  break;
        }
      }
    (void) eml_lostcontact(contact);
    }
  cycle++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to do TLS transaction with a	*/
/*	binded port. Stop its loop under	*/
/*	CTRL-C					*/
/*						*/
/************************************************/
static void dotls(BPTR *bptr,char *tlshost,char *tlsport)

{
#define	OPTS	"modtst.c:dotls,"
#define	MXWAIT	5	/*wait 5 sec for contact	*/

u_long	cycle;
u_int pool;

cycle=0;
while (sigterm==false) {
  CTTYP *contact;
  int socker;
  int phase;
  int proceed;

  contact=(CTTYP *)0;
  socker=0;
  phase=0;
  proceed=true;
  while (proceed==true) {
    switch (phase) {
      case   0	:	/*opening a connection		*/
	if ((contact=eml_opencontact(pro_smtp,tlshost,tlsport,MXWAIT,&socker))==(CTTYP *)0) {
	  proceed=false;
	  sigterm=true;
	  }
	break;
      case  1	: 	/*is contact established?	*/
	if (contact->established==false) {
          (void) clm_alert(0,"Caller process %05d contact to <%s> trouble <%s>",
			      getpid(),tlshost,contact->info);
	  phase=999;	/*trouble trouble		*/
	  }
	break;
      case  2	: {	/*send a NOOP command		*/
	register int i;

	for (i=0;i<10;i++)
	  if (eml_sendcmd(contact,CMDOK,"%s","NOOP")!=CMDOK)
	    break;
        }
	break;
      default	:	/*closing and looping back	*/
	contact=eml_closecontact(contact);
	cycle++;
	phase=-1;
        if (sigterm==true)
	  proceed=false;
	break;
      }
    phase++;
    }
  }
if ((pool=ipc_getpool(bptr,getppid()))!=(u_long)0) {
  char *msg;
  TIMESPEC timeout;
  char *msgref;

  (void) asprintf(&msg,"%d %lu",getpid(),cycle);
  timeout.tv_sec=1;
  timeout.tv_nsec=0;
  if ((msgref=ipc_postmsg(bptr,&timeout,pool,msg))!=(char *)0) {
    (void) free(msgref);
    }
  (void) free(msg);
  }
else
  (void) clm_alert(0,"Caller process %05d can't report to master",getpid());
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to check if tls handling is done*/
/*	properly.				*/
/*						*/
/************************************************/
static int tsttls()

{
#define TLSDURE	10
#define	MXTLS	 4

static BININF bindinfo[]={{pro_smtp,"127.1.0.1","10025",3},
			  {pro_smtp,"127.2.0.1","10025",2}};

BININF *blst[]={&(bindinfo[0]),&(bindinfo[1]),(BININF *)0};
int success;
u_long totcpttls;
u_long start;
u_long stop;
MXCTYP *mxc;
LSTTYP **look;
pid_t child[MXTLS];
int i;
int received;
int phase;
int proceed;

success=false;
totcpttls=(u_long)0;
start=(u_long)0;
stop=(u_long)0;
mxc=(MXCTYP *)calloc(1,sizeof(MXCTYP));
look=(LSTTYP **)0;
received=0;
sigint=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*setting the share memory	*/
      (void) clm_settrap(true);
      if ((mxc->bptr=ipc_getbox(MXBOX,10))==(BPTR *)0) {
        (void) clm_alert(0,"Unable to assigne shared memory");
	proceed=false;
	}
    case   1	:	/*preparing binding		*/
      if ((look=pro_getlst(look,blst))==(LSTTYP **)0) {
	phase=999;
        }
      break;
    case   2	:	/*starting listening process	*/
      if ((mxc->ckrpool=ipc_openpool(mxc->bptr))==(u_long)0)
 	(void) clm_core_dump("Unable to create the main pool");
      if (pro_listen(false,mxc,look,tlsaction)==false)
        phase=999;
      break;
    case   3	:	/*tacking care msg exchange	*/
      if (ipc_subscribe(mxc->bptr,(u_long)0,true)==false)
 	(void) clm_core_dump("Unable to subscribe main process to IPC");
      break;
    case   4	:	
      (void) clm_alert(0,"Starting TLS Exchange test");
      break;
    case   5	: 	/*starting writing process	*/
      for (i=0;i<MXTLS;i++) {
	switch (child[i]=fork()) {
	  case  -1	:
            (void) clm_alert(0,"Unable to fork TLS process");
	    break;
	  case   0	:
	    (void) dotls(mxc->bptr,bindinfo[i%2].ip,bindinfo[i%2].port);
	    (void) exit(0);
	    break;
	  default	:
	    break;
          }
        }
      break;
    case   6	:	/*waiting manual Halt		*/
      start=clm_getmillisec();
      if (debug==0)
        (void) clm_alert(0,"Wait '%d' seconds for test completion",TLSDURE);
      else
        (void) clm_alert(0,"Enter CTRL-C to stop test");
      while (sigint==false) {
        (void) sleep(1);
	if (debug==0) {
	  if ((clm_getmillisec()-start)>(TLSDURE*1000)) {
	    sigint=true;
	    }
	  }
	}
      for (i=0;i<MXTLS;i++) {
        (void) kill(child[i],SIGTERM);
	}
      stop=clm_getmillisec();
      (void) clm_alert(0,"Ending test and starting to collect data");
      break;
    case   7	:	/*stopping writing process	*/
      while (received<MXTLS) {
        char *got;
	char *msgref;
	TIMESPEC timer;
	u_long millisec;

        timer.tv_sec=20;
        timer.tv_nsec=0;
        if ((got=ipc_getpost(mxc->bptr,&timer,&msgref,&millisec))!=(char *)0) {
	  int call;
	  u_long cpttls;

	  call=0;
	  cpttls=(u_long)0;
	  if (sscanf(got,"%d %lu",&call,&cpttls)==2) {
	    totcpttls+=cpttls;
	    }
	  (void) free(msgref);
	  (void) free(got);
	  received++;
	  }
	else
	  break;
	}
      (void) clm_alert(0,"Found a total of '%lu' TLS sessions",totcpttls);
      if (totcpttls>0) {
	float duration;

	duration=((float)stop-start)/1000.0;
        (void) clm_alert(0,"TLS speed %6.2f sessions/sec for %8.3f sec",
			   (float)totcpttls/duration,duration);
	}
      break;
    case   8	:	/*stopping listening process	*/
      (void) dev_mulclose(look);
      look=pro_freelst(look);
      break;
    case   9	:	/*unsubscribing from pool	*/
      (void) ipc_subscribe(mxc->bptr,(u_long)0,false);
      mxc->ckrpool=ipc_closepool(mxc->bptr,mxc->ckrpool);
      break;
    default	:	/*SAFE Guard			*/
      mxc->bptr=ipc_dropbox(mxc->bptr);
      (void) clm_settrap(false);
      (void) clm_alert(0,"TLS Exchange test completed");
      success=true;
      proceed=false;	
      break;
    }
  phase++;
  }
(void) free(mxc);
return success;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to check if config file are	*/
/*	propely parsed by configurator module.	*/
/*						*/
/************************************************/
static int tstcnf()

{
const char *empty;
CNFTYP *cnf;
SQLFILE *sql;
int result;
int phase;
int proceed;

empty=(const char *)0;
(void) clm_settrap(true);
(void) clm_alert(0,"Testing config");
(void) clm_setjournal(APJRL,d_log,appname);
cnf=(CNFTYP *)0;
sql=(SQLFILE *)0;
result=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*opening data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to open database","modtst.c:tstcnf");
	proceed=false;	/*TROUBLE		*/
	}
      break;
    case  1	:	/*loading the config parameters	*/
      if ((cnf=cnf_chkconfig(sql,cnf))==(CNFTYP *)0) {
	proceed=false;
	}
      break;
    case  2	:	/*dumping config information	*/
      (void) cnf_dumpconfig(stdout,cnf);
      break;
    case  3	:	/*tracing result		*/
      if (result==0)
        (void) clm_alert(0,"Probing config successful");
      break;
    default	:	/*SAFE guard			*/
      cnf=cnf_freeconfig(cnf);
      sql=sql_close(sql);
      proceed=false;
      break;
    }
  if (result<0) {
    (void) clm_alert(0,"Testing config error within phase '%d' (test abort)",phase);
    proceed=false;
    }
  phase++;
  }
(void) clm_alert(0,"Config test completed");
(void) jrl_closejournals();
(void) clm_settrap(false);
return true;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to handle module function	*/
/*	testing.				*/
/*						*/
/************************************************/
int tst_module(char *rqst)

{
typedef	enum	{
	mod_all,	/*all module		*/
	mod_jrl,	/*journalling		*/
	mod_ipc,	/*message exchanges	*/
	mod_tls,	/*TLS exchanges		*/
	mod_cnf,	/*Config parsing check	*/
	mod_unk		/*unknown request	*/
	}MODENU;

static VOCTYP voctst[]={
	{mod_all,"All","\tTesting all module"},
	{mod_jrl,"Journal","Testing journalling module"},
	{mod_ipc,"Messaging","Testing Message exchange module"},
	{mod_tls,"TLS","Testing TLS (over TCP/IP) message Exchange"},
	{mod_cnf,"CNF","Testing config parsing et refresh"},
	{mod_unk,(char *)0,"Module testing available command are:"}
	};

int success;
VOCTYP *voc;

success=false;
if (rqst==(char *)0)
  rqst="All";
if ((voc=clm_getvoca(voctst,rqst))!=(VOCTYP *)0) {
  switch (voc->code) {
    case mod_all	:	/*testing all module	*/
      success=true;
      (void) fprintf(stdout,"Testing all modules\n");
      success&=tstjournal();
      success&=tstipc();
      success&=tsttls();
      break;
    case mod_jrl	:	/*testing journal module*/
      (void) tstjournal();
      break;
    case mod_ipc	:	/*testing journal module*/
      success=tstipc();
      break;
    case mod_tls	:	/*testing tls module*/
      success=tsttls();
      break;
    case mod_cnf	:	/*testing config module*/
      success=tstcnf();
      break;
    case mod_unk	:	/*unknown request	*/
      (void) fprintf(stdout,"%s\n",(char *)voctst[voc->code].ptr);
      voc=voctst;
      while (voc->key!=(char *)0) {
        (void) fprintf(stdout,"\t'%s':\t%s\n",voc->key,(char *)voc->ptr);
	voc++;
        }
      break;
    }
  }
return success;
}
