/************************************************/
/*						*/
/*	UNIRLY					*/
/*	Define relaying authorisation module.	*/
/*	All procedure involved to allow an Email*/
/*	relaying are located in this module	*/
/*						*/
/************************************************/
#include	<sys/stat.h>
#include	<arpa/inet.h>
#include	<errno.h>
#include	<string.h>
#
#include	"lowtyp.h"
#include	"basafn.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unirly.h"

#define	RLYENV	"RELAY"	/*Relay file ENV var	*/
#define	RLYLST	"alwaysrelay:popbsmtp"

typedef	struct	{
	u_int groupid;	/*relayed IP group	*/
	AFNTYP *afnnum;	/*IP number		*/
	char *from;	/*originator email addr	*/
	time_t upto;	/*maximun relayable	*/
	}RLYENT;

typedef	struct	{
	char *fullname;	/*file name to scan	*/
	time_t lastscan;/*file lastscan		*/
	RLYENT **relais;/*relais list info	*/
	}RLYFILE;

typedef	struct	{
	time_t lastscan;/*last general scan	*/
	RLYFILE **infos;/*file scanned contents	*/
	}RLYINF;
/*

*/
/************************************************/
/*						*/
/*	Procedure to scan a relay str format	*/
/*	en convert it to a relay entry.		*/
/*						*/
/************************************************/
static RLYENT *scanentry(char *relaydata)

{
#define	IP4MASK	32
#define	OPSC	"unirly.c:scanentry,"

RLYENT *entry;
u_int groupid;
time_t upto;
char *from;
char ip[50];
int mask;
int num;
char *strip;
char *strfrom;
char *strdate;
char *strtime;

entry=(RLYENT *)0;
groupid=0;
upto=(time_t)0;
from=(char *)0;
strip=calloc(strlen(relaydata)+1,sizeof(char));
strfrom=calloc(strlen(relaydata)+1,sizeof(char));
strdate=calloc(strlen(relaydata)+1,sizeof(char));
strtime=calloc(strlen(relaydata)+1,sizeof(char));
mask=IP4MASK;
num=sscanf(relaydata,"%u %s %s %s %s",&groupid,strip,strfrom,strdate,strtime);
switch (num)	{	/*order matter	*/
  case	5	:	/*get time	*/
  case	4	:	/*get date	*/
    upto=clm_datetimesysasc(strdate,strtime);
    /*NO BREAK!	*/
  case	3	:	/*get from	*/
    from=strdup(strfrom);
    /*NO BREAK!	*/
  case	2	:	/*get IP	*/
    if (sscanf(strip,"%49[^/]/%d",ip,&mask)>=1) {
      AFNTYP *afnnum;

      if ((afnnum=afn_getipnum(ip))==(AFNTYP *)0) {
	(void) clm_alert(0,"%s Unable to convert IP <%s> from <%s> (error=<%s>)",
			    OPSC,ip,relaydata,strerror(errno));
        break;
	}
      (void) afn_maskip(afnnum,mask);
      entry=(RLYENT *)calloc(1,sizeof(RLYENT));
      entry->groupid=groupid;
      entry->afnnum=afnnum;
      entry->from=from;
      entry->upto=upto;
      break;
    case  1	:	/*only group ID?*/
    case  0	:	/*trouble!	*/
      (void) clm_alert(1,"%s Unable to scan <%s>",OPSC,relaydata);
      break;
    }
  }
(void) free(strtime);
(void) free(strdate);
(void) free(strfrom);
(void) free(strip);
return entry;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to scan a relay info and	*/
/*	store it within a relay structure.	*/
/*	Format is:				*/
/*	Group_num,IP_num,[domain,[expire_time]]	*/
/*						*/
/************************************************/
static void updateinfstr(RLYFILE *info,char **relaydata)

{
info->relais=rly_freerelais(info->relais);
if (relaydata!=(char **)0) {
  char **ptr;

  for (ptr=relaydata;*ptr!=(char *)0;ptr++) {
    RLYENT *entry;
   
    if ((entry=scanentry(*ptr))!=(RLYENT *)0) 
      info->relais=(RLYENT **)vec_addlstlst((void **)info->relais,(void *)entry);
    else
      (void) clm_alert(0,"%s format error found in relay table <%s>",
			 "unirly.c:updateinfstr,",info->fullname);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	relais array.				*/
/*						*/
/************************************************/
static RLYENT **freerelais(RLYENT **relais)

{
if (relais!=(RLYENT **)0) {
  int i;

  for (i=0;relais[i]!=(RLYENT *)0;i++) {
    relais[i]->afnnum=afn_freeipnum(relais[i]->afnnum);
    relais[i]->from=clm_freestr(relais[i]->from);
    (void) free(relais[i]);
    }
  (void) free(relais);
  relais=(RLYENT **)0;
  }
return relais;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a minimal RLYINF	*/
/*	structure contents.			*/
/*						*/
/************************************************/
static RLYINF *preprelay()

{
RLYINF *relais;
const char *ptr;
char *list;
char **split;
int phase;
int proceed;

relais=(RLYINF *)0;
ptr=(const char *)0;
list=(char *)0;
split=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting file list	*/
      if ((ptr=getenv(RLYENV))==(const char *)0) {
        ptr=RLYLST;
	}
      list=strdup(ptr); 
      break;
    case 1	:	/*do we have a list	*/
      if (strlen(list)==0)
        phase=999;	/*no list! get out quick*/
      break;
    case 2	:	/*preparing the struct	*/
      relais=(RLYINF *)calloc(1,sizeof(RLYINF));
      relais->lastscan=(time_t)0;
      break;
    case 3	:	/*list of file to scan	*/
      if ((split=clm_splitstring(list,':'))!=(char **)0) {
        register const char *etcpath;
	int i;
        
	etcpath=clm_appdir(d_etc);
	for (i=0;split[i]!=(char *)0;i++) {
	  RLYFILE *info;
          
	  info=(RLYFILE *)calloc(1,sizeof(RLYFILE));
	  (void) asprintf(&(info->fullname),"%s%s",etcpath,split[i]);
          info->lastscan=(time_t)0;
	  relais->infos=(RLYFILE **)vec_addlstlst((void **)relais->infos,(void *)info);
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      split=(char **)vec_freelstlst((void **)split,VECFREE);
      (void) free(list);
      proceed=false;
      break;
    }
  phase++;
  }
return relais;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update one relais info list*/
/*						*/
/************************************************/
static RLYFILE *updateinfos(RLYFILE *infos)

{
#define	OPUP	"unirly.c:updateinfos,"

#define	MXB	1024
#define	IP4MASK	32

time_t updtime;
FILE *fichier;
char buffer[MXB];
int phase;
int proceed;

updtime=(time_t)0;
fichier=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Was the file updated?	*/
      if (infos->fullname!=(char *)0) {
	struct stat bufstat;

	if (stat(infos->fullname,&bufstat)<0) {
	  (void) clm_alert(1,"%s Unable to get stat for <%s>, (error=<%s>)",
			     OPUP,infos->fullname,strerror(errno));
          infos->relais=freerelais(infos->relais);
	  proceed=false;/*trouble trouble	*/
	  }
	else {
	  if (bufstat.st_mtime<=infos->lastscan) {
	    updtime=bufstat.st_mtime;
	    phase=999;	/*no need to go further	*/
	    }
	  }
        }
      break;
    case 1	:	/*cleaning infos record	*/
      infos->relais=freerelais(infos->relais);
      break;
    case 2	:	/*opening file		*/
      if ((fichier=fopen(infos->fullname,"r"))==(FILE *)0) {
	(void) clm_alert(1,"%s Unable to open file <%s>, (error=<%s>)",
			    OPUP,infos->fullname,strerror(errno));
	phase=999;	/*no need to go further	*/
	}
      break;
    case 3	:	/*reading file		*/
      (void) strcpy(buffer,"1 ");
      while (clm_getstr(fichier,buffer+2,sizeof(buffer)-3,'#')!=(char *)0) {
	RLYENT *entry;

        if ((entry=scanentry(buffer))!=(RLYENT *)0) 
	  infos->relais=(RLYENT **)vec_addlstlst((void **)infos->relais,(void *)entry);
	}
      break;
    case 4	:	/*closing file		*/
      (void) fclose(fichier);
      infos->lastscan=updtime;
      break;
    default	:	/*SAFE GUARD		*/
      proceed=false;
      break;
    }
  phase++;
  }
return infos;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a specific	*/
/*	IP is with the relayable list		*/
/*						*/
/************************************************/
int rly_isinrelay(RLYTYP *relais,char *peerip,time_t curtime)

{
register int isin;

isin=false;
if ((relais!=(RLYTYP *)0)&&(((RLYINF *)relais)->infos!=(RLYFILE **)0)) {
  AFNTYP *afnnum;

  if ((afnnum=afn_getipnum(peerip))==(AFNTYP *)0) {
    (void) clm_alert(0,"%s Unable to scan IP <%s> (error=<%s>)","unirly.c:rly_isinrelay,",peerip,strerror(errno));
    }
  else {
    int i;

    for (i=0;(isin==false)&&((RLYINF *)relais)->infos[i]!=(RLYFILE *)0;i++) {
      int j;

      if (((RLYINF *)relais)->infos[i]->relais==(RLYENT **)0)
	continue;
      for (j=0;((RLYINF *)relais)->infos[i]->relais[j]!=(RLYENT *)0;j++) {
        RLYENT *ptr;

        ptr=((RLYINF *)relais)->infos[i]->relais[j];
        if (afn_cmpipnum(ptr->afnnum,afnnum,ptr->afnnum->afnmask)==false)
          continue;
        if ((ptr->upto>0)&&(curtime>ptr->upto))
          continue;
        isin=true;
        break;
	}
      }
    afnnum=afn_freeipnum(afnnum);
    }
  }
return isin;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a specific	*/
/*	originator is relayable.		*/
/*						*/
/************************************************/
static RLYENUM isrelayable(RLYFILE *relay,char *from,AFNTYP *afnnum,time_t curtime,int *groupid)

{
RLYENUM relayable;

*groupid=0;
relayable=rly_norelay;
if ((relay!=(RLYFILE *)0)&&(relay->relais!=(RLYENT **)0)) {
  int i;

  for (i=0;(relayable==rly_norelay)&&(relay->relais[i]!=(RLYENT *)0);i++) {
    RLYENT *ptr;
    int phase;
    int proceed;

    ptr=relay->relais[i];
    phase=0;
    proceed=true;
    while (proceed==true) {
      
      switch (phase) {
	case 0		:	/*checking IP	*/
	  if (afn_cmpipnum(ptr->afnnum,afnnum,ptr->afnnum->afnmask)==false)
	    proceed=false;	/*no luck	*/
	  break;
	case 1		:	/*checking date	*/
	  if ((ptr->upto>0)&&(curtime>ptr->upto))
	    proceed=false;	/*no luck	*/
	  break;
	case 2		:	/*checking from	*/
	  if ((ptr->from!=(char *)0)&&(strcmp(ptr->from,"*")==0)) {
	    *groupid=ptr->groupid;
	    relayable=rly_forcedrelay;
            proceed=false;	/*we must relay!*/
	    }
	  break;
	case 3		:	/*any domain?	*/
	  if ((ptr->from!=(char *)0)&&(strcmp(ptr->from,"-")==0))
	    phase=999;		/*bingo!	*/
	  break;
	case 4		:	/*checking from	*/
	  if ((ptr->from!=(char *)0)&&(strcmp(ptr->from,from)!=0))
	    proceed=false;	/*no luck	*/
	  break;
        default		:	/*SAFE guard	*/
	  *groupid=ptr->groupid;
	  relayable=rly_smartrelay;	/*Bingo!*/
          proceed=false;
	  break;
	}
      phase++;
      }
    }
  }
return relayable;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the date of the	*/
/*	most recent update done among the	*/
/*	relaying file				*/
/*						*/
/************************************************/
time_t rly_lastupd(RLYTYP **relais)

{
#define	OPLA	"unirly.c:rly_lastupd,"

time_t lastupd;
RLYINF *ptrrel;
int phase;
int proceed;

lastupd=(time_t)0;
ptrrel=(RLYINF *)*relais;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*already a struct ready*/
      if (ptrrel==(RLYINF *)0) {
	*relais=preprelay();
        if (*relais==(RLYINF *)0)
	  proceed=false;
	}
      break;
    case 1	:	/*check last update	*/
      ptrrel=(RLYINF *)*relais;
      if (ptrrel->infos!=(RLYFILE **)0) {
        int i;
	
	for (i=0;ptrrel->infos[i]!=(RLYFILE *)0;i++) {
	  RLYFILE *ptr;

	  ptr=ptrrel->infos[i];	
      	  if (ptr->fullname!=(char *)0) {
	    struct stat bufstat;

	    if (stat(ptr->fullname,&bufstat)==0) {
	      if (bufstat.st_mtime>lastupd) 
	        lastupd=bufstat.st_mtime;
	      }
	    }
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return lastupd;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to fee memory used by relay	*/
/*	data structure.				*/
/*	is allow to relay an Email		*/
/*						*/
/************************************************/
RLYTYP *rly_freerelais(RLYTYP *relais)

{
if (relais!=(RLYTYP *)0) {
  RLYINF *ptr;

  ptr=(RLYINF *)relais;
  if (ptr->infos!=(RLYFILE **)0) {
    int i;

    for (i=0;ptr->infos[i]!=(RLYFILE *)0;i++) {
      ptr->infos[i]->relais=freerelais(ptr->infos[i]->relais);
      ptr->infos[i]->fullname=clm_freestr(ptr->infos[i]->fullname);
      (void) free(ptr->infos[i]);
      }
    (void) free(ptr->infos);
    }
  (void) free(ptr);
  relais=(RLYTYP *)0;
  }
return relais;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a originator	*/
/*	is allow to relay an Email		*/
/*						*/
/************************************************/
RLYENUM rly_isrelayable(RLYTYP *relais,char *from,char *peerip,int *groupid)

{
#define	OPLY	"unirly.c:rly_isrelayable,"

RLYENUM relayable;
AFNTYP *afnnum;
time_t curtime;
int phase;
int proceed;

relayable=rly_norelay;
afnnum=(AFNTYP *)0;
curtime=time((time_t *)0);
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case 0	:	/*check if relais	*/
      if (relais==(RLYINF *)0) 
	proceed=false;	/*no relais data avail	*/
      break;
    case 1	:	/*converting IP address	*/
      if ((afnnum=afn_getipnum(peerip))==(AFNTYP *)0) {
	(void) clm_alert(0,"%s Unable to scan IP <%s> (error=<%s>)",OPLY,peerip,strerror(errno));
	proceed=false;
	}
      break;
    case 2	:	/*scanning for relay	*/
      if (((RLYINF *)relais)->infos!=(RLYFILE **)0) {
        int i;
	
	for (i=0;((RLYINF *)relais)->infos[i]!=(RLYFILE *)0;i++) {
	  relayable=isrelayable(((RLYINF *)relais)->infos[i],from,afnnum,curtime,groupid);
	  if (relayable!=rly_norelay)
	    break;
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      afnnum=afn_freeipnum(afnnum);
      proceed=false;
      break;
    }
  phase++;
  }
return relayable;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a relay list within	*/
/*	the relayable structure.		*/
/*						*/
/************************************************/
RLYTYP *rly_addrelais(RLYTYP *relais,char *rlyname,char **rlyliste)

{
RLYINF *ptr;
int phase;
int proceed;

ptr=(RLYINF *)(relais);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we have a relayname*/
      if ((rlyname==(char *)0)||(strlen(rlyname)==0))
	phase=999;	/*no need to go further	*/
      break;
    case 1	:	/*do we have relay file?*/
      if (ptr==(RLYINF *)0) {
	ptr=preprelay();
      if (ptr==(RLYINF *)0) {
        ptr=(RLYINF *)calloc(1,sizeof(RLYINF));
        ptr->lastscan=(time_t)0;
	}
      break;
    case 2	:	/*already a list loaded	*/
      if (ptr->infos!=(RLYFILE **)0) {
	int i;

	for (i=0;ptr->infos[i]!=(RLYFILE *)0;i++) 
	  if (strcmp(ptr->infos[i]->fullname,rlyname)==0) {
	    (void) updateinfstr(ptr->infos[i],rlyliste);
	    phase++;	/*no need to add info	*/
	    break;
	    }
	  }
	}
      break;
    case 3	:	/*creating needed info	*/
      if (ptr!=(RLYINF *)0) {	/*always	*/
	RLYFILE *info;
          
	info=(RLYFILE *)calloc(1,sizeof(RLYFILE));
	info->fullname=strdup(rlyname);
	(void) updateinfstr(info,rlyliste);
	ptr->infos=(RLYFILE **)vec_addlstlst((void **)ptr->infos,(void *)info);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return (RLYTYP *)ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to load the relay structure	*/
/*						*/
/************************************************/
RLYTYP *rly_loadrlyfile(RLYTYP *relais)

{
time_t curtime;
int phase;
int proceed;

curtime=time((time_t *)0);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*already a struct ready*/
      if (relais==(RLYINF *)0) {
	relais=preprelay();
        if (relais==(RLYINF *)0)
	  proceed=false;
	}
      break;
    case 1	:	/*lets refresh entries	*/
      if (((RLYINF *)relais)->infos!=(RLYFILE **)0) {
        int i;
	
	for (i=0;((RLYINF *)relais)->infos[i]!=(RLYFILE *)0;i++) {
	  ((RLYINF *)relais)->infos[i]=updateinfos(((RLYINF *)relais)->infos[i]);
	  }
	}
      ((RLYINF *)relais)->lastscan=curtime;
      break;
    default	:	/*SAFE guard		*/
      if (((RLYINF *)relais)->infos!=(RLYFILE **)0)
      proceed=false;
      break;
    }
  phase++;
  }
return (RLYTYP *)relais;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare string used to	*/
/*	display relai table contents.		*/
/*						*/
/************************************************/
char **rly_showrelais(RLYTYP *relais)

{
char **strrly;

strrly=(char **)0;
if (((RLYINF *)relais!=(RLYINF *)0)&&(((RLYINF *)relais)->infos!=(RLYFILE **)0)) {
  RLYFILE **ptr;

  for (ptr=((RLYINF *)relais)->infos;*ptr!=(RLYFILE *)0;ptr++) {
    if ((*ptr)->relais!=(RLYENT **)0) {
      char *str;
      RLYENT **ent;

      (void) asprintf(&str,"Relay file: %s",(*ptr)->fullname);
      strrly=(char **)vec_addlstlst((void **)strrly,(void *)str);
      for (ent=(*ptr)->relais;*ent!=(RLYENT *)0;ent++) {
	char *afnstr;
	char *stripnum;
	char *strfrom;
	char *strdate;

	afnstr=afn_stripnum((*ent)->afnnum);
	(void) asprintf(&stripnum,"%s/%d",afnstr,(*ent)->afnnum->afnmask);
	if ((*ent)->from==(char *)0)
	  strfrom=strdup("-");
	else
	  strfrom=strdup((*ent)->from);
	if ((*ent)->upto==(time_t)0)
	  strdate=strdup("always");
	else
	  (void) asprintf(&strdate,"up to %s",clm_ascsysdatetime((*ent)->upto));
	(void) asprintf(&str,"\t%03u  %*s%s\t%s",(*ent)->groupid,-50,stripnum,strfrom,strdate);
	(void) free(strdate);
	(void) free(strfrom);
	(void) free(stripnum);
	afnstr=clm_freestr(afnstr);
        strrly=(char **)vec_addlstlst((void **)strrly,(void *)str);
	}	
      }
    }
  }
return strrly;
}
